"""
Fixed target data collection
"""
from __future__ import annotations

import logging
import os
import shutil
import sys
import time
from datetime import datetime
from pathlib import Path
from time import sleep
from typing import Dict, List

import numpy as np
from dodal.beamlines import i24
from dodal.devices.i24.pmac import PMAC

from mx_bluesky.I24.serial import log
from mx_bluesky.I24.serial.dcid import DCID
from mx_bluesky.I24.serial.fixed_target.ft_utils import (
    ChipType,
    MappingType,
    PumpProbeSetting,
)
from mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_StartUp_py3v1 import (
    get_format,
    scrape_parameter_file,
)
from mx_bluesky.I24.serial.parameters import SSXType
from mx_bluesky.I24.serial.parameters.constants import LITEMAP_PATH, PARAM_FILE_PATH_FT
from mx_bluesky.I24.serial.setup_beamline import caget, cagetstring, caput, pv
from mx_bluesky.I24.serial.setup_beamline import setup_beamline as sup
from mx_bluesky.I24.serial.setup_beamline.setup_detector import get_detector_type
from mx_bluesky.I24.serial.write_nexus import call_nexgen

logger = logging.getLogger("I24ssx.fixed_target")

usage = "%(prog)s [options]"


def setup_logging():
    # Log should now change name daily.
    logfile = time.strftime("i24fixedtarget_%d%B%y.log").lower()
    log.config(logfile)


def flush_print(text):
    sys.stdout.write(str(text))
    sys.stdout.flush()


def copy_files_to_data_location(
    dest_dir: Path | str,
    param_path: Path = PARAM_FILE_PATH_FT,
    map_file: Path = LITEMAP_PATH,
    map_type: str = "1",
):
    if not isinstance(dest_dir, Path):
        dest_dir = Path(dest_dir)
    dest_dir.mkdir(parents=True, exist_ok=True)
    shutil.copy2(param_path / "parameters.txt", dest_dir / "parameters.txt")
    if map_type == MappingType.Lite:
        shutil.copy2(map_file / "currentchip.map", dest_dir / "currentchip.map")


@log.log_on_entry
def get_chip_prog_values(
    chip_type,
    pump_repeat,
    pumpexptime,
    pumpdelay,
    prepumpexptime,
    exptime=16,
    n_exposures=1,
):
    if chip_type in [ChipType.Oxford, ChipType.OxfordInner, ChipType.Minichip]:
        logger.info("This is an Oxford chip %s" % chip_type)
        # '1' = 'Oxford ' = [8, 8, 20, 20, 0.125, 3.175, 3.175]
        (
            xblocks,
            yblocks,
            x_num_steps,
            y_num_steps,
            w2w,
            b2b_horz,
            b2b_vert,
        ) = get_format(chip_type)
        x_step_size = w2w
        y_step_size = w2w
        x_block_size = ((x_num_steps - 1) * w2w) + b2b_horz
        y_block_size = ((y_num_steps - 1) * w2w) + b2b_vert

    elif chip_type == ChipType.Custom:
        # This is set by the user in the edm screen
        # The chip format might change every time and is read from PVs.
        logger.info("This is a Custom Chip")
        x_num_steps = caget(pv.me14e_gp6)
        y_num_steps = caget(pv.me14e_gp7)
        x_step_size = caget(pv.me14e_gp8)
        y_step_size = caget(pv.me14e_gp99)
        xblocks = 1
        yblocks = 1
        x_block_size = 0  # placeholder
        y_block_size = 0  # placeholder
    else:
        logger.warning("Unknown chip_type, chip_type = %s" % chip_type)

    # this is where p variables for fast laser expts will be set
    if pump_repeat in [
        PumpProbeSetting.NoPP,
        PumpProbeSetting.Short1,
        PumpProbeSetting.Short2,
    ]:
        pump_repeat_pvar = 0
    elif pump_repeat == PumpProbeSetting.Repeat1:
        pump_repeat_pvar = 1
    elif pump_repeat == PumpProbeSetting.Repeat2:
        pump_repeat_pvar = 2
    elif pump_repeat == PumpProbeSetting.Repeat3:
        pump_repeat_pvar = 3
    elif pump_repeat == PumpProbeSetting.Repeat5:
        pump_repeat_pvar = 5
    elif pump_repeat == PumpProbeSetting.Repeat10:
        pump_repeat_pvar = 10
    else:
        logger.warning("Unknown pump_repeat, pump_repeat = %s" % pump_repeat)

    logger.info("Pump repeat is %s, PVAR set to %s" % (pump_repeat, pump_repeat_pvar))

    if pump_repeat == PumpProbeSetting.Short2:
        pump_in_probe = 1
    else:
        pump_in_probe = 0

    logger.info("pump_in_probe set to %s" % pump_in_probe)

    chip_dict = {
        "X_NUM_STEPS": [11, x_num_steps],
        "Y_NUM_STEPS": [12, y_num_steps],
        "X_STEP_SIZE": [13, x_step_size],
        "Y_STEP_SIZE": [14, y_step_size],
        "DWELL_TIME": [15, exptime],
        "X_START": [16, 0],
        "Y_START": [17, 0],
        "Z_START": [18, 0],
        "X_NUM_BLOCKS": [20, xblocks],
        "Y_NUM_BLOCKS": [21, yblocks],
        "X_BLOCK_SIZE": [24, x_block_size],
        "Y_BLOCK_SIZE": [25, y_block_size],
        "COLTYPE": [26, 41],
        "N_EXPOSURES": [30, n_exposures],
        "PUMP_REPEAT": [32, pump_repeat_pvar],
        "LASER_DWELL": [34, pumpexptime],
        "LASERTWO_DWELL": [35, prepumpexptime],
        "LASER_DELAY": [37, pumpdelay],
        "PUMP_IN_PROBE": [38, pump_in_probe],
    }

    chip_dict["DWELL_TIME"][1] = 1000 * float(exptime)
    chip_dict["LASER_DWELL"][1] = 1000 * float(pumpexptime)
    chip_dict["LASERTWO_DWELL"][1] = 1000 * float(prepumpexptime)
    chip_dict["LASER_DELAY"][1] = 1000 * float(pumpdelay)

    return chip_dict


@log.log_on_entry
def load_motion_program_data(
    pmac: PMAC, motion_program_dict: Dict[str, List], map_type: int, pump_repeat: int
):
    logger.info("Loading motion program data for chip.")
    logger.info("Pump_repeat is %s" % pump_repeat)
    if pump_repeat == PumpProbeSetting.NoPP:
        if map_type == MappingType.NoMap:
            prefix = 11
            logger.info(
                "Map type is %s, setting program prefix to %s" % (map_type, prefix)
            )
        elif map_type == MappingType.Lite:
            prefix = 12
        elif map_type == MappingType.Full:
            prefix = 13
        else:
            logger.warning("Unknown Map Type, map_type = %s" % map_type)
            return
    elif pump_repeat in [pp.value for pp in PumpProbeSetting if pp != 0]:
        # Pump setting chosen
        prefix = 14
        logger.info("Setting program prefix to %s" % prefix)
        pmac.pmac_string.set("P1439=0")
        if bool(caget(pv.me14e_gp111)) is True:
            logger.info("Checker pattern setting enabled.")
            pmac.pmac_string.set("P1439=1")
    else:
        logger.warning("Unknown Pump repeat, pump_repeat = %s" % pump_repeat)
        return

    logger.info("Set PMAC_STRING pv.")
    for key in sorted(motion_program_dict.keys()):
        v = motion_program_dict[key]
        pvar_base = prefix * 100
        pvar = pvar_base + v[0]
        value = str(v[1])
        s = "P%s=%s" % (str(pvar), str(value))
        logger.info("%s \t %s" % (key, s))
        pmac.pmac_string.set(s)
        sleep(0.02)
    sleep(0.2)


@log.log_on_entry
def get_prog_num(chip_type: int, map_type: int, pump_repeat: int):
    logger.info("Get Program Number")
    if pump_repeat == PumpProbeSetting.NoPP:
        if chip_type in [ChipType.Oxford, ChipType.OxfordInner]:
            logger.info("Pump_repeat: %s \tOxford Chip: %s" % (pump_repeat, chip_type))
            if map_type == MappingType.NoMap:
                logger.info("Map type 0 = None")
                logger.info("Program number: 11")
                return 11
            elif map_type == MappingType.Lite:
                logger.info("Map type 1 = Mapping Lite")
                logger.info("Program number: 12")
                return 12
            elif map_type == MappingType.Full:
                logger.info("Map type 1 = Full Mapping")
                logger.info("Program number: 13")  # once fixed return 13
                msg = "Mapping Type FULL is broken as of 11.09.17"
                logger.error(msg)
                raise ValueError(msg)
            else:
                logger.debug("Unknown Mapping Type; map_type = %s" % map_type)
                return 0
        elif chip_type == ChipType.Custom:
            logger.info("Pump_repeat: %s \tCustom Chip: %s" % (pump_repeat, chip_type))
            logger.info("Program number: 11")
            return 11
        elif chip_type == ChipType.Minichip:
            logger.info(
                "Pump_repeat: %s \tMini Oxford Chip: %s" % (pump_repeat, chip_type)
            )
            logger.info("Program number: 11")
            return 11
        else:
            logger.debug("Unknown chip_type, chip_tpe = = %s" % chip_type)
            return 0
    elif pump_repeat in [pp.value for pp in PumpProbeSetting if pp != 0]:
        logger.info("Pump_repeat: %s \t Chip Type: %s" % (pump_repeat, chip_type))
        logger.info("Map Type = Mapping Lite with Pump Probe")
        logger.info("Program number: 14")
        return 14
    else:
        logger.warning("Unknown pump_repeat, pump_repeat = = %s" % pump_repeat)
        return 0


@log.log_on_entry
def datasetsizei24():
    # Calculates how many images will be collected based on map type and N repeats
    logger.info("Calculate total number of images expected in data collection.")
    (
        chip_name,
        visit,
        sub_dir,
        n_exposures,
        chip_type,
        map_type,
        pump_repeat,
        pumpexptime,
        pumpdelay,
        exptime,
        dcdetdist,
        prepumpexptime,
        det_type,
    ) = scrape_parameter_file()

    if map_type == MappingType.NoMap:
        if chip_type == ChipType.Custom:
            total_numb_imgs = int(int(caget(pv.me14e_gp6)) * int(caget(pv.me14e_gp7)))
            logger.info(
                "Map type: None \tCustom chip \tNumber of images %s" % total_numb_imgs
            )
        else:
            chip_format = get_format(chip_type)[:4]
            total_numb_imgs = np.prod(chip_format)
            logger.info(
                "Map type: None \tOxford chip %s \tNumber of images %s"
                % (chip_type, total_numb_imgs)
            )

    elif map_type == MappingType.Lite:
        logger.info("Using Mapping Lite on chip type %s" % chip_type)
        chip_format = get_format(chip_type)[2:4]
        block_count = 0
        with open(LITEMAP_PATH / "currentchip.map", "r") as f:
            for line in f.readlines():
                entry = line.split()
                if entry[2] == "1":
                    block_count += 1

        logger.info("Block count=%s" % block_count)
        logger.info("Chip format=%s" % chip_format)

        n_exposures = int(caget(pv.me14e_gp3))
        logger.info("Number of exposures=%s" % n_exposures)

        total_numb_imgs = np.prod(chip_format) * block_count * n_exposures
        logger.info("Calculated number of images: %s" % total_numb_imgs)

    elif map_type == MappingType.Full:
        logger.error("Not Set Up For Full Mapping")
        raise ValueError("The beamline is currently not set for Full Mapping.")

    else:
        logger.warning("Unknown Map Type, map_type = %s" % map_type)
        raise ValueError("Unknown map type")

    logger.info("Set PV to calculated number of images.")
    caput(pv.me14e_gp10, total_numb_imgs)

    return total_numb_imgs


@log.log_on_entry
def start_i24():
    """Returns a tuple of (start_time, dcid)"""
    logger.info("Start I24 data collection.")
    start_time = datetime.now()
    logger.info("Collection start time %s" % start_time.ctime())

    (
        chip_name,
        visit,
        sub_dir,
        n_exposures,
        chip_type,
        map_type,
        pump_repeat,
        pumpexptime,
        pumpdelay,
        exptime,
        dcdetdist,
        prepumpexptime,
        det_type,
    ) = scrape_parameter_file()

    logger.debug("Set up beamline")
    sup.beamline("collect")
    sup.beamline("quickshot", [dcdetdist])
    logger.debug("Set up beamline DONE")

    total_numb_imgs = datasetsizei24()

    filepath = visit + sub_dir
    filename = chip_name

    logger.debug("Acquire Region")

    num_gates = int(total_numb_imgs) / int(n_exposures)

    logger.info("Total number of images: %d" % total_numb_imgs)
    logger.info("Number of exposures: %s" % n_exposures)
    logger.info("Number of gates (=Total images/N exposures): %.4f" % num_gates)

    if det_type == "pilatus":
        logger.info("Using Pilatus detector")
        logger.info("Fastchip Pilatus setup: filepath %s" % filepath)
        logger.info("Fastchip Pilatus setup: filepath %s" % filename)
        logger.info("Fastchip Pilatus setup: number of images %d" % total_numb_imgs)
        logger.info("Fastchip Pilatus setup: exposure time %s" % exptime)

        sup.pilatus("fastchip", [filepath, filename, total_numb_imgs, exptime])

        # DCID process depends on detector PVs being set up already
        logger.debug("Start DCID process")
        dcid = DCID(
            emit_errors=False,
            ssx_type=SSXType.FIXED,
            visit=Path(visit).name,
            image_dir=filepath,
            start_time=start_time,
            num_images=total_numb_imgs,
            exposure_time=exptime,
            detector=det_type,
            shots_per_position=int(n_exposures),
            pump_exposure_time=float(pumpexptime),
            pump_delay=float(pumpdelay),
            pump_status=int(pump_repeat),
        )

        logger.debug("Arm Pilatus. Arm Zebra.")
        sup.zebra1("fastchip-pilatus", [num_gates, n_exposures, exptime])
        caput(pv.pilat_acquire, "1")  # Arm pilatus
        caput(pv.zebra1_pc_arm, "1")  # Arm zebra fastchip-pilatus
        caput(pv.pilat_filename, filename)
        time.sleep(1.5)

    elif det_type == "eiger":
        logger.info("Using Eiger detector")

        logger.warning(
            """TEMPORARY HACK!
            Running a Single image pilatus data collection to create directory."""
        )
        num_imgs = 1
        sup.pilatus("quickshot-internaltrig", [filepath, filename, num_imgs, exptime])
        logger.debug("Sleep 2s waiting for pilatus to arm")
        sleep(2)
        sleep(0.5)
        caput(pv.pilat_acquire, "0")  # Disarm pilatus
        sleep(0.5)
        caput(pv.pilat_acquire, "1")  # Arm pilatus
        logger.debug("Pilatus data collection DONE")
        sup.pilatus("return to normal")
        logger.info("Pilatus back to normal. Single image pilatus data collection DONE")

        logger.info("Triggered Eiger setup: filepath %s" % filepath)
        logger.info("Triggered Eiger setup: filename %s" % filename)
        logger.info("Triggered Eiger setup: number of images %d" % total_numb_imgs)
        logger.info("Triggered Eiger setup: exposure time %s" % exptime)

        sup.eiger("triggered", [filepath, filename, total_numb_imgs, exptime])

        # DCID process depends on detector PVs being set up already
        logger.debug("Start DCID process")
        dcid = DCID(
            emit_errors=False,
            ssx_type=SSXType.FIXED,
            visit=Path(visit).name,
            image_dir=filepath,
            start_time=start_time,
            num_images=total_numb_imgs,
            exposure_time=exptime,
            detector=det_type,
        )

        logger.debug("Arm Zebra.")
        sup.zebra1("fastchip-eiger", [num_gates, n_exposures, exptime])
        caput(pv.zebra1_pc_arm, "1")  # Arm zebra fastchip-eiger

        time.sleep(1.5)

    else:
        msg = "Unknown Detector Type, det_type = %s" % det_type
        logger.error(msg)
        raise ValueError(msg)

    # Open the hutch shutter

    caput("BL24I-PS-SHTR-01:CON", "Reset")
    logger.debug("Reset, then sleep for 1s")
    sleep(1.0)
    caput("BL24I-PS-SHTR-01:CON", "Open")
    logger.debug(" Open, then sleep for 2s")
    sleep(2.0)

    return start_time.ctime(), dcid


@log.log_on_entry
def finish_i24(chip_prog_dict, start_time):
    det_type = get_detector_type()
    logger.info("Finish I24 data collection with %s detector." % det_type)

    (
        chip_name,
        visit,
        sub_dir,
        n_exposures,
        chip_type,
        map_type,
        pump_repeat,
        pumpexptime,
        pumpdelay,
        exptime,
        dcdetdist,
        prepumpexptime,
        det_type,
    ) = scrape_parameter_file()

    total_numb_imgs = datasetsizei24()
    filepath = visit + sub_dir
    filename = chip_name
    transmission = (float(caget(pv.pilat_filtertrasm)),)
    wavelength = float(caget(pv.dcm_lambda))

    if det_type == "pilatus":
        logger.info("Finish I24 Pilatus")
        filename = filename + "_" + caget(pv.pilat_filenum)
        logger.debug("Close the fast shutter.")
        caput(pv.zebra1_soft_in_b1, "No")
        logger.debug("Disarm the zebra.")
        caput(pv.zebra1_pc_arm_out, "0")
        sup.zebra1("return-to-normal")
        sup.pilatus("return-to-normal")
        sleep(0.2)
    elif det_type == "eiger":
        logger.info("Finish I24 Eiger")
        logger.debug("Close the fast shutter.")
        caput(pv.zebra1_soft_in_b1, "No")
        logger.debug("Disarm the zebra.")
        caput(pv.zebra1_pc_arm_out, "0")
        sup.zebra1("return-to-normal")
        sup.eiger("return-to-normal")
        filename = cagetstring(pv.eiger_ODfilenameRBV)

    # Detector independent moves
    logger.info("Move chip back to home position by setting PMAC_STRING pv.")
    caput(pv.me14e_pmac_str, "!x0y0z0")
    logger.debug("Closing shutter")
    caput("BL24I-PS-SHTR-01:CON", "Close")

    end_time = time.ctime()
    logger.info("Collection end time %s" % end_time)

    # Copy parameter file and eventual chip map to collection directory
    copy_files_to_data_location(Path(visit + sub_dir), map_type=map_type)

    # Write a record of what was collected to the processing directory
    userlog_path = visit + "processing/" + sub_dir + "/"
    userlog_fid = filename + "_parameters.txt"
    logger.debug("Write a user log in %s" % userlog_path)

    os.makedirs(userlog_path, exist_ok=True)

    with open(userlog_path + userlog_fid, "w") as f:
        f.write("Fixed Target Data Collection Parameters\n")
        f.write("Data directory \t%s\n" % filepath)
        f.write("Filename \t%s\n" % filename)
        f.write("Shots per pos \t%s\n" % n_exposures)
        f.write("Total N images \t%s\n" % total_numb_imgs)
        f.write("Exposure time \t%s\n" % exptime)
        f.write("Det distance \t%s\n" % dcdetdist)
        f.write("Transmission \t%s\n" % transmission)
        f.write("Wavelength \t%s\n" % wavelength)
        f.write("Detector type \t%s\n" % det_type)
        f.write("Pump status \t%s\n" % pump_repeat)
        f.write("Pump exp time \t%s\n" % pumpexptime)
        f.write("Pump delay \t%s\n" % pumpdelay)

    sleep(0.5)

    return end_time


def main():
    # Dodal devices
    pmac = i24.pmac()
    # ABORT BUTTON
    logger.info("Running a chip collection on I24")
    caput(pv.me14e_gp9, 0)

    logger.info("Getting parameters from file.")
    (
        chip_name,
        visit,
        sub_dir,
        n_exposures,
        chip_type,
        map_type,
        pump_repeat,
        pumpexptime,
        pumpdelay,
        exptime,
        dcdetdist,
        prepumpexptime,
        det_type,
    ) = scrape_parameter_file()

    logger.info("Chip name is %s" % chip_name)
    logger.info("visit = %s" % visit)
    logger.info("sub_dir = %s" % sub_dir)
    logger.info("n_exposures = %s" % n_exposures)
    logger.info("chip_type = %s" % chip_type)
    logger.info("map_type = %s" % map_type)
    logger.info("dcdetdist = %s" % dcdetdist)
    logger.info("exptime = %s" % exptime)
    logger.info("pump_repeat = %s" % pump_repeat)
    logger.info("pumpexptime = %s" % pumpexptime)
    logger.info("pumpdelay = %s" % pumpdelay)
    logger.info("prepumpexptime = %s" % prepumpexptime)
    logger.info("Getting Program Dictionary")

    # If alignment type is Oxford inner it is still an Oxford type chip
    if chip_type == ChipType.OxfordInner:
        logger.debug("Change chip type Oxford Inner to Oxford.")
        chip_type = ChipType.Oxford

    chip_prog_dict = get_chip_prog_values(
        chip_type,
        pump_repeat,
        pumpexptime,
        pumpdelay,
        prepumpexptime,
        exptime=exptime,
        n_exposures=n_exposures,
    )
    logger.info("Loading Motion Program Data")
    load_motion_program_data(pmac, chip_prog_dict, map_type, pump_repeat)

    start_time, dcid = start_i24()

    logger.info("Moving to Start")
    caput(pv.me14e_pmac_str, "!x0y0z0")
    sleep(2.0)

    prog_num = get_prog_num(chip_type, map_type, pump_repeat)

    # Now ready for data collection. Open fast shutter
    logger.debug("Opening fast shutter.")
    caput(pv.zebra1_soft_in_b1, "1")  # Open fast shutter (zebra gate)

    logger.info("Run PMAC with program number %d" % prog_num)
    logger.info("pmac str = &2b%dr" % prog_num)
    caput(pv.me14e_pmac_str, "&2b%dr" % prog_num)
    sleep(1.0)

    # Kick off the StartOfCollect script
    logger.debug("Notify DCID of the start of the collection.")
    dcid.notify_start()

    param_file_tuple = scrape_parameter_file()
    tot_num_imgs = datasetsizei24()
    if det_type == "eiger":
        logger.debug("Start nexus writing service.")
        call_nexgen(
            chip_prog_dict,
            start_time,
            param_file_tuple,
            total_numb_imgs=tot_num_imgs,
        )

    logger.info("Data Collection running")

    aborted = False
    timeout_time = time.time() + tot_num_imgs * float(exptime) + 60

    # me14e_gp9 is the ABORT button
    if int(caget(pv.me14e_gp9)) == 0:
        i = 0
        text_list = ["|", "/", "-", "\\"]
        while True:
            line_of_text = "\r\t\t\t Waiting   " + 30 * ("%s" % text_list[i % 4])
            flush_print(line_of_text)
            sleep(0.5)
            i += 1
            if int(caget(pv.me14e_gp9)) != 0:
                aborted = True
                logger.warning("Data Collection Aborted")
                caput(pv.me14e_pmac_str, "A")
                sleep(1.0)
                caput(pv.me14e_pmac_str, "P2401=0")
                break
            elif int(caget(pv.me14e_scanstatus)) == 0:
                # As soon as me14e_scanstatus is set to 0, exit.
                # Epics checks the geobrick and updates this PV every s or so.
                # Once the collection is done, it will be set to 0.
                print(caget(pv.me14e_scanstatus))
                logger.warning("Data Collection Finished")
                break
            elif time.time() >= timeout_time:
                aborted = True
                logger.warning(
                    """
                    Something went wrong and data collection timed out. Aborting.
                    """
                )
                caput(pv.me14e_pmac_str, "A")
                sleep(1.0)
                caput(pv.me14e_pmac_str, "P2401=0")
                break
    else:
        aborted = True
        logger.info("Data Collection ended due to GP 9 not equalling 0")

    logger.debug("Closing fast shutter")
    caput(pv.zebra1_soft_in_b1, "No")  # Close the fast shutter
    sleep(2.0)

    if det_type == "pilatus":
        logger.debug("Pilatus Acquire STOP")
        sleep(0.5)
        caput(pv.pilat_acquire, 0)
    elif det_type == "eiger":
        logger.debug("Eiger Acquire STOP")
        sleep(0.5)
        caput(pv.eiger_acquire, 0)
        caput(pv.eiger_ODcapture, "Done")

    end_time = finish_i24(chip_prog_dict, start_time)
    dcid.collection_complete(end_time, aborted=aborted)
    logger.debug("Notify DCID of end of collection.")
    dcid.notify_end()

    logger.info("Quick summary of settings")
    logger.info("Chip name = %s sub_dir = %s" % (chip_name, sub_dir))
    logger.info("Start Time = % s" % start_time)
    logger.info("End Time = %s" % end_time)


if __name__ == "__main__":
    setup_logging()

    main()



from __future__ import annotations

import functools
import logging
import logging.config
from os import environ
from pathlib import Path
from typing import Optional

# Logging set up
logging.getLogger("I24ssx").addHandler(logging.NullHandler())

logging_config = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "class": "logging.Formatter",
            "format": "%(message)s",
        }
    },
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "default",
            "stream": "ext://sys.stdout",
        }
    },
    "loggers": {
        "I24ssx": {
            "handlers": ["console"],
            "level": "DEBUG",
            "propagate": True,
        }
    },
}

logging.config.dictConfig(logging_config)


def _get_logging_file_path() -> Path:
    """Get the path to write the artemis log files to.
    If on a beamline, this will be written to the according area depending on the
    BEAMLINE envrionment variable. If no envrionment variable is found it will default
    it to the tmp/dev directory.
    Returns:
        logging_path (Path): Path to the log file for the file handler to write to.
    """
    beamline: Optional[str] = environ.get("BEAMLINE")
    logging_path: Path

    if beamline:
        logging_path = Path("/dls_sw/" + beamline + "/logs/serial/")
    else:
        logging_path = Path("./tmp/logs/")

    try:
        Path(logging_path).mkdir(parents=True, exist_ok=True)
    except OSError:
        # Until https://github.com/DiamondLightSource/mx_bluesky/issues/45 is fixed
        # Logs could also go to the current visit directory, but not always possible
        # when testing
        logging_path = Path("~/serial_logs/").expanduser().resolve()
        Path(logging_path).mkdir(parents=True, exist_ok=True)
    return logging_path


def config(logfile: str | None = None, write_mode: str = "a", delayed: bool = False):
    """
    Configure the logging.

    Args:
        logfile (str, optional): Filename for logfile. If passed, create a file handler\
            for the logger to write to file the log output. Defaults to None.
        write_mode (str, optional): String indicating writing mode for the output \
            .log file. Defaults to "a".
    """
    logger = logging.getLogger("I24ssx")
    if logfile:
        logs = _get_logging_file_path() / logfile
        fileFormatter = logging.Formatter(
            "%(asctime)s %(levelname)s: \t(%(name)s) %(message)s",
            datefmt="%d-%m-%Y %I:%M:%S",
        )
        FH = logging.FileHandler(logs, mode=write_mode, encoding="utf-8", delay=delayed)
        FH.setLevel(logging.DEBUG)
        FH.setFormatter(fileFormatter)
        logger.addHandler(FH)


def log_on_entry(func):
    logger = logging.getLogger("I24ssx")

    @functools.wraps(func)
    def decorator(*args, **kwargs):
        name = func.__name__
        logger.debug("Running %s " % name)
        return func(*args, **kwargs)

    return decorator



from unittest.mock import ANY, MagicMock, call, patch

import cv2 as cv
import pytest

from mx_bluesky.I24.serial.fixed_target.i24ssx_moveonclick import (
    onMouse,
    update_ui,
)


@pytest.mark.parametrize(
    "beam_position, expected_1J, expected_2J",
    [
        ((15, 10), "#1J:-90", "#2J:-60"),
        ((100, 150), "#1J:-600", "#2J:-900"),
        ((475, 309), "#1J:-2850", "#2J:-1854"),
        ((638, 392), "#1J:-3828", "#2J:-2352"),
    ],
)
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_moveonclick.i24.pmac")
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_moveonclick.get_beam_centre")
def test_onMouse_gets_beam_position_and_sends_correct_str(
    fake_get_beam_pos,
    fake_pmac,
    beam_position,
    expected_1J,
    expected_2J,
):
    fake_get_beam_pos.side_effect = [beam_position]
    fake_pmac.pmac_string = MagicMock()
    onMouse(cv.EVENT_LBUTTONUP, 0, 0, "", param=fake_pmac)
    fake_pmac.pmac_string.assert_has_calls(
        [
            call.set(expected_1J),
            call.set(expected_2J),
        ]
    )


@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_moveonclick.cv")
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_moveonclick.get_beam_centre")
def test_update_ui_uses_correct_beam_centre_for_ellipse(fake_beam_pos, fake_cv):
    mock_frame = MagicMock()
    fake_beam_pos.side_effect = [(15, 10)]
    update_ui(mock_frame)
    fake_cv.ellipse.assert_called_once()
    fake_cv.ellipse.assert_has_calls(
        [call(ANY, (15, 10), (12, 8), 0.0, 0.0, 360, (0, 255, 255), thickness=2)]
    )



"""
Define a mapping between the edm screens/IOC enum inputs for the general purpose PVs and
the map/chip/pump settings.
The enum values should not be changed unless they are also modified in the drop down
menu in the edm screen, as their order should always match.
New ones may be added if needed in the future.
"""
from enum import IntEnum


class MappingType(IntEnum):
    NoMap = 0
    Lite = 1
    Full = 2


class ChipType(IntEnum):
    Oxford = 0
    OxfordInner = 1
    Custom = 2
    Minichip = 3  # Mini oxford, 1 city block only


class PumpProbeSetting(IntEnum):
    NoPP = 0
    Short1 = 1
    Short2 = 2
    Repeat1 = 3
    Repeat2 = 4
    Repeat3 = 5
    Repeat5 = 6
    Repeat10 = 7



from functools import partial
from unittest.mock import patch

import pytest
from bluesky.run_engine import RunEngine
from dodal.beamlines import i24
from dodal.devices.i24.I24_detector_motion import DetectorMotion
from ophyd.status import Status

from mx_bluesky.I24.serial.parameters.constants import SSXType
from mx_bluesky.I24.serial.setup_beamline import Eiger, Pilatus
from mx_bluesky.I24.serial.setup_beamline.setup_detector import (
    DetRequest,
    get_detector_type,
    setup_detector_stage,
)


@pytest.fixture
def fake_detector_motion() -> DetectorMotion:
    detector_motion = i24.detector_motion(fake_with_ophyd_sim=True)
    detector_motion.y.user_setpoint._use_limits = False
    detector_motion.z.user_setpoint._use_limits = False

    def mock_set(motor, val):
        motor.user_readback.sim_put(val)
        return Status(done=True, success=True)

    def patch_motor(motor):
        return patch.object(motor, "set", partial(mock_set, motor))

    with patch_motor(detector_motion.y), patch_motor(detector_motion.z):
        yield detector_motion


@patch("mx_bluesky.I24.serial.setup_beamline.setup_detector.caget")
def test_get_detector_type(fake_caget):
    fake_caget.return_value = -22
    assert get_detector_type().name == "eiger"


@patch("mx_bluesky.I24.serial.setup_beamline.setup_detector.caget")
def test_get_detector_type_finds_pilatus(fake_caget):
    fake_caget.return_value = 566
    assert get_detector_type().name == "pilatus"


@patch("mx_bluesky.I24.serial.setup_beamline.setup_detector.caget")
def test_setup_detector_stage(fake_caget, fake_detector_motion):
    RE = RunEngine()

    fake_caget.return_value = DetRequest.eiger.value
    RE(setup_detector_stage(fake_detector_motion, SSXType.FIXED))
    assert fake_detector_motion.y.user_readback.get() == Eiger.det_y_target

    fake_caget.return_value = DetRequest.pilatus.value
    RE(setup_detector_stage(fake_detector_motion, SSXType.EXTRUDER))
    assert fake_detector_motion.y.user_readback.get() == Pilatus.det_y_target



"""
Move on click gui for fixed targets at I24
Robin Owen 12 Jan 2021
"""
import logging

import cv2 as cv
from dodal.beamlines import i24
from dodal.devices.oav.oav_detector import OAV

from mx_bluesky.I24.serial.fixed_target import i24ssx_Chip_Manager_py3v1 as manager
from mx_bluesky.I24.serial.parameters.constants import OAV1_CAM

logger = logging.getLogger("I24ssx.moveonclick")

# Set scale.
# TODO See https://github.com/DiamondLightSource/mx_bluesky/issues/44
zoomcalibrator = 6  # 8 seems to work well for zoom 2


def _get_beam_centre(oav: OAV):
    """Extract the beam centre x/y positions from the display.configuration file.

    Args:
        oav (OAV): the OAV device.
    """
    return oav.parameters.beam_centre_i, oav.parameters.beam_centre_j


# TODO In the future, this should be done automatically in the OAV device
# See https://github.com/DiamondLightSource/dodal/issues/224
def get_beam_centre():
    # Get I24 oav device from dodal
    oav = i24.oav()

    beamX, beamY = _get_beam_centre(oav)
    return beamX, beamY


# Register clicks and move chip stages
def onMouse(event, x, y, flags, param):
    pmac = param
    beamX, beamY = get_beam_centre()
    if event == cv.EVENT_LBUTTONUP:
        logger.info("Clicked X and Y %s %s" % (x, y))
        xmove = -1 * (beamX - x) * zoomcalibrator
        ymove = -1 * (beamY - y) * zoomcalibrator
        logger.info("Moving X and Y %s %s" % (xmove, ymove))
        xmovepmacstring = "#1J:" + str(xmove)
        ymovepmacstring = "#2J:" + str(ymove)
        pmac.pmac_string.set(xmovepmacstring)
        pmac.pmac_string.set(ymovepmacstring)


def update_ui(frame):
    # Get beam x and y values
    beamX, beamY = get_beam_centre()

    # Overlay text and beam centre
    cv.ellipse(
        frame, (beamX, beamY), (12, 8), 0.0, 0.0, 360, (0, 255, 255), thickness=2
    )
    # putText(frame,'text',bottomLeftCornerOfText, font, fontScale, fontColor, thickness, lineType)
    cv.putText(
        frame,
        "Key bindings",
        (20, 40),
        cv.FONT_HERSHEY_COMPLEX_SMALL,
        1,
        (0, 255, 255),
        1,
        1,
    )
    cv.putText(
        frame,
        "Q / A : go to / set as f0",
        (25, 70),
        cv.FONT_HERSHEY_COMPLEX_SMALL,
        0.8,
        (0, 255, 255),
        1,
        1,
    )
    cv.putText(
        frame,
        "W / S : go to / set as f1",
        (25, 90),
        cv.FONT_HERSHEY_COMPLEX_SMALL,
        0.8,
        (0, 255, 255),
        1,
        1,
    )
    cv.putText(
        frame,
        "E / D : go to / set as f2",
        (25, 110),
        cv.FONT_HERSHEY_COMPLEX_SMALL,
        0.8,
        (0, 255, 255),
        1,
        1,
    )
    cv.putText(
        frame,
        "I / O : in /out of focus",
        (25, 130),
        cv.FONT_HERSHEY_COMPLEX_SMALL,
        0.8,
        (0, 255, 255),
        1,
        1,
    )
    cv.putText(
        frame,
        "C : Create CS",
        (25, 150),
        cv.FONT_HERSHEY_COMPLEX_SMALL,
        0.8,
        (0, 255, 255),
        1,
        1,
    )
    cv.putText(
        frame,
        "esc : close window",
        (25, 170),
        cv.FONT_HERSHEY_COMPLEX_SMALL,
        0.8,
        (0, 255, 255),
        1,
        1,
    )
    cv.imshow("OAV1view", frame)


def start_viewer(oav1: str = OAV1_CAM):
    # Get PMAC device
    pmac = i24.pmac()
    # Create a video caputure from OAV1
    cap = cv.VideoCapture(oav1)

    # Create window named OAV1view and set onmouse to this
    cv.namedWindow("OAV1view")
    cv.setMouseCallback("OAV1view", onMouse, param=pmac)  # type: ignore

    logger.info("Showing camera feed. Press escape to close")
    # Read captured video and store them in success and frame
    success, frame = cap.read()

    # Loop until escape key is pressed. Keyboard shortcuts here
    while success:
        success, frame = cap.read()

        update_ui(frame)

        k = cv.waitKey(1)
        if k == 113:  # Q
            manager.moveto("zero")
        if k == 119:  # W
            manager.moveto("f1")
        if k == 101:  # E
            manager.moveto("f2")
        if k == 97:  # A
            pmac.pmac_string.set(r"\#1hmz\#2hmz\#3hmz")
            print("Current position set as origin")
        if k == 115:  # S
            manager.fiducial(1)
        if k == 100:  # D
            manager.fiducial(2)
        if k == 99:  # C
            manager.cs_maker()
        if k == 98:  # B
            manager.block_check()  # doesn't work well for blockcheck as image doesn't update
        if k == 104:  # H
            pmac.pmac_string.set("#2J:-10")
        if k == 110:  # N
            pmac.pmac_string.set("#2J:10")
        if k == 109:  # M
            pmac.pmac_string.set("#1J:-10")
        if k == 98:  # B
            pmac.pmac_string.set("#1J:10")
        if k == 105:  # I
            pmac.pmac_string.set("#3J:-150")
        if k == 111:  # O
            pmac.pmac_string.set("#3J:150")
        if k == 117:  # U
            pmac.pmac_string.set("#3J:-1000")
        if k == 112:  # P
            pmac.pmac_string.set("#3J:1000")
        if k == 0x1B:  # esc
            cv.destroyWindow("OAV1view")
            print("Pressed escape. Closing window")
            break

    # Clear cameraCapture instance
    cap.release()


if __name__ == "__main__":
    start_viewer()



from enum import Enum
from pathlib import Path


class SSXType(Enum):
    FIXED = "Serial Fixed"
    EXTRUDER = "Serial Jet"


OAV_CONFIG_FILES = {
    "zoom_params_file": "/dls_sw/i24/software/gda/config/xml/jCameraManZoomLevels.xml",
    "oav_config_json": "/dls_sw/i24/software/daq_configuration/json/OAVCentring.json",
    "display_config": "/dls_sw/i24/software/gda_versions/var/display.configuration",
}
OAV1_CAM = "http://bl24i-di-serv-01.diamond.ac.uk:8080/OAV1.mjpg.mjpg"

PARAM_FILE_PATH = Path("src/mx_bluesky/I24/serial/parameters").expanduser().resolve()
PARAM_FILE_PATH_FT = (
    Path("src/mx_bluesky/I24/serial/parameters/fixed_target").expanduser().resolve()
)
LITEMAP_PATH = (
    Path("src/mx_bluesky/I24/serial/parameters/fixed_target/litemaps")
    .expanduser()
    .resolve()
)
FULLMAP_PATH = (
    Path("src/mx_bluesky/I24/serial/parameters/fixed_target/fullmaps")
    .expanduser()
    .resolve()
)
PVAR_FILE_PATH = (
    Path("src/mx_bluesky/I24/serial/parameters/fixed_target/pvar_files")
    .expanduser()
    .resolve()
)
HEADER_FILES_PATH = Path("/dls_sw/i24/scripts/fastchips/").expanduser().resolve()
CS_FILES_PATH = (
    Path("src/mx_bluesky/I24/serial/parameters/fixed_target/cs").expanduser().resolve()
)



import json
from unittest.mock import ANY, MagicMock, call, mock_open, patch

import pytest

from mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1 import (
    cs_maker,
    cs_reset,
    moveto,
    parse_args_and_run_parsed_function,
    pumpprobe_calc,
    scrape_mtr_directions,
    scrape_mtr_fiducials,
)

mtr_dir_str = """#Some words
mtr1_dir=1
mtr2_dir=-1
mtr3_dir=-1"""

fiducial_1_str = """MTR RBV RAW Corr f_value
MTR1 0 0 1 0
MTR2 1 -1 -1 1
MTR3 0 0 -1 0"""

cs_json = '{"scalex":1, "scaley":2, "scalez":3, "skew":-0.5, "Sx_dir":1, "Sy_dir":-1, "Sz_dir":0}'


@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caput")
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caget")
def test_moveto_oxford_origin(fake_caget, fake_caput):
    fake_caget.return_value = 0
    moveto("origin")
    assert fake_caget.call_count == 1
    assert fake_caput.call_count == 2


@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caput")
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caget")
def test_moveto_oxford_inner_f1(fake_caget, fake_caput):
    fake_caget.return_value = 1
    moveto("f1")
    assert fake_caget.call_count == 1
    assert fake_caput.call_count == 2


@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caput")
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caget")
def test_moveto_chip_unknown(fake_caget, fake_caput):
    fake_caget.return_value = 4
    moveto("zero")
    assert fake_caget.call_count == 1
    assert fake_caput.call_count == 1


@patch(
    "mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.open",
    mock_open(read_data=mtr_dir_str),
)
def test_scrape_mtr_directions():
    res = scrape_mtr_directions()
    assert len(res) == 3
    assert res == (1.0, -1.0, -1.0)


@patch(
    "mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.open",
    mock_open(read_data=fiducial_1_str),
)
def test_scrape_mtr_fiducials():
    res = scrape_mtr_fiducials(1)
    assert len(res) == 3
    assert res == (0.0, 1.0, 0.0)


@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caput")
def test_cs_reset(fake_caput):
    cs_reset()
    assert fake_caput.call_count == 4


@patch(
    "mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.open",
    mock_open(read_data='{"a":11, "b":12,}'),
)
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caput")
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caget")
@patch(
    "mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.scrape_mtr_directions"
)
@patch(
    "mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.scrape_mtr_fiducials"
)
def test_cs_maker_raises_error_for_invalid_json(
    fake_fid,
    fake_dir,
    fake_caget,
    fake_caput,
):
    fake_dir.return_value = (1, 1, 1)
    fake_fid.return_value = (0, 0, 0)
    with pytest.raises(json.JSONDecodeError):
        cs_maker()


@patch(
    "mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.open",
    mock_open(read_data='{"scalex":11, "skew":12}'),
)
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caput")
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caget")
@patch(
    "mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.scrape_mtr_directions"
)
@patch(
    "mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.scrape_mtr_fiducials"
)
def test_cs_maker_raises_error_for_missing_key_in_json(
    fake_fid,
    fake_dir,
    fake_caget,
    fake_caput,
):
    fake_dir.return_value = (1, 1, 1)
    fake_fid.return_value = (0, 0, 0)
    with pytest.raises(KeyError):
        cs_maker()


@patch(
    "mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.open",
    mock_open(read_data=cs_json),
)
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caput")
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caget")
@patch(
    "mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.scrape_mtr_directions"
)
@patch(
    "mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.scrape_mtr_fiducials"
)
def test_cs_maker_raises_error_for_wrong_direction_in_json(
    fake_fid,
    fake_dir,
    fake_caget,
    fake_caput,
):
    fake_dir.return_value = (1, 1, 1)
    fake_fid.return_value = (0, 0, 0)
    with pytest.raises(ValueError):
        cs_maker()


@patch(
    "mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.fiducial",
    autospec=True,
)
def test_arg_parser_runs_fiducial_function_as_expected(mock_fiducial: MagicMock):
    parse_args_and_run_parsed_function(["fiducial", "2"])
    mock_fiducial.assert_called_once_with(2)


@patch(
    "mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.define_current_chip",
    autospec=True,
)
def test_arg_parser_runs_define_current_chip_function_as_expected(
    mock_define_current_chip: MagicMock,
):
    parse_args_and_run_parsed_function(["define_current_chip", "chip_id"])
    mock_define_current_chip.assert_called_once_with("chip_id")


@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caput")
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Manager_py3v1.caget")
def test_pumpprobe_calc(fake_caget, fake_caput):
    fake_caget.side_effect = [0.01, 0.005]
    pumpprobe_calc()
    assert fake_caget.call_count == 2
    assert fake_caput.call_count == 5
    fake_caput.assert_has_calls(
        [
            call(ANY, 0.62),
            call(ANY, 1.24),
            call(ANY, 1.86),
            call(ANY, 3.1),
            call(ANY, 6.2),
        ]
    )



from unittest.mock import patch

import pytest

from mx_bluesky.I24.serial.setup_beamline import setup_beamline


@patch("mx_bluesky.I24.serial.setup_beamline.setup_beamline.caput")
def test_beamline_collect(fake_caput):
    setup_beamline.beamline("collect")
    assert fake_caput.call_count == 4


def test_beamline_raises_error_if_quickshot_and_no_args_list():
    with pytest.raises(TypeError):
        setup_beamline.beamline("quickshot")


@patch("mx_bluesky.I24.serial.setup_beamline.setup_beamline.caput")
@patch("mx_bluesky.I24.serial.setup_beamline.setup_beamline.caget")
def test_beamline_quickshot(fake_caget, fake_caput):
    fake_caget.return_value = 100
    setup_beamline.beamline("quickshot", ["100"])
    assert fake_caput.call_count == 1
    assert fake_caget.call_count == 2


@patch("mx_bluesky.I24.serial.setup_beamline.setup_beamline.caput")
@patch("mx_bluesky.I24.serial.setup_beamline.setup_beamline.caget")
def test_pilatus_raises_error_if_fastchip_and_no_args_list(fake_caget, fake_caput):
    with pytest.raises(TypeError):
        setup_beamline.pilatus("fastchip")


@patch("mx_bluesky.I24.serial.setup_beamline.setup_beamline.caput")
@patch("mx_bluesky.I24.serial.setup_beamline.setup_beamline.caget")
def test_pilatus_quickshot(fake_caget, fake_caput):
    setup_beamline.pilatus("quickshot", ["", "", 1, 0.1])
    assert fake_caput.call_count == 12
    assert fake_caget.call_count == 2


@patch("mx_bluesky.I24.serial.setup_beamline.setup_beamline.caput")
@patch("mx_bluesky.I24.serial.setup_beamline.setup_beamline.caget")
def test_eiger_raises_error_if_quickshot_and_no_args_list(fake_caget, fake_caput):
    with pytest.raises(TypeError):
        setup_beamline.eiger("quickshot")


@patch("mx_bluesky.I24.serial.setup_beamline.setup_beamline.caput")
@patch("mx_bluesky.I24.serial.setup_beamline.setup_beamline.caget")
def test_eiger_quickshot(fake_caget, fake_caput):
    setup_beamline.eiger("quickshot", ["", "", "1", "0.1"])
    assert fake_caput.call_count == 32


@patch("mx_bluesky.I24.serial.setup_beamline.setup_beamline.caput")
def test_zebra1_return_to_normal(fake_caput):
    setup_beamline.zebra1("return-to-normal")
    assert fake_caput.call_count == 20


@patch("mx_bluesky.I24.serial.setup_beamline.setup_beamline.caput")
def test_zebra1_quickshot(fake_caput):
    setup_beamline.zebra1("quickshot", [0, 1])
    assert fake_caput.call_count == 7


@patch("mx_bluesky.I24.serial.setup_beamline.setup_beamline.caput")
def test_zebra1_fastchip_pilatus(fake_caput):
    setup_beamline.zebra1("fastchip-pilatus", [1, 1, 0.1])
    assert fake_caput.call_count == 12


@patch("mx_bluesky.I24.serial.setup_beamline.setup_beamline.caput")
def test_zebra1_fastchip_eiger(fake_caput):
    setup_beamline.zebra1("fastchip-eiger", [1, 1, 0.1])
    assert fake_caput.call_count == 12



import logging
from unittest.mock import patch

import pytest

from mx_bluesky.I24.serial import log


@pytest.fixture
def dummy_logger():
    logger = logging.getLogger("I24ssx")
    yield logger


@patch("mx_bluesky.I24.serial.log.environ")
@patch("mx_bluesky.I24.serial.log.Path.mkdir")
def test_logging_file_path(mock_dir, mock_environ):
    mock_environ.get.return_value = None
    log_path = log._get_logging_file_path()
    assert mock_dir.call_count == 1
    assert log_path.as_posix() == "tmp/logs"


@patch("mx_bluesky.I24.serial.log.environ")
@patch("mx_bluesky.I24.serial.log.Path.mkdir")
def test_logging_file_path_on_beamline(mock_dir, mock_environ):
    mock_environ.get.return_value = "i24"
    log_path = log._get_logging_file_path()
    assert mock_dir.call_count == 1
    assert log_path.as_posix() == "/dls_sw/i24/logs/serial"


def test_basic_logging_config(dummy_logger):
    assert dummy_logger.hasHandlers() is True
    assert len(dummy_logger.handlers) == 1
    assert dummy_logger.handlers[0].level == logging.DEBUG


@patch("mx_bluesky.I24.serial.log.Path.mkdir")
def test_logging_config_with_filehandler(mock_dir, dummy_logger):
    log.config("dummy.log", delayed=True)
    assert len(dummy_logger.handlers) == 2
    assert mock_dir.call_count == 1
    assert dummy_logger.handlers[1].level == logging.DEBUG
    # Clear FileHandler to avoid other tests failing if it is kept open
    dummy_logger.removeHandler(dummy_logger.handlers[1])



from unittest.mock import MagicMock, call, mock_open, patch

import pytest

from mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Collect_py3v1 import (
    datasetsizei24,
    get_chip_prog_values,
    get_prog_num,
    load_motion_program_data,
    start_i24,
)

params = {
    "chip_name": "chip",
    "visit": "foo",
    "sub_dir": "bar",
    "n_exposures": 1,
    "chip_type": 0,
    "map_type": 1,
    "pump_repeat": 0,
    "pumpexptime": 0,
    "pumpdelay": 0,
    "exptime": 0.01,
    "dcdetdist": 100,
    "prepumpexptime": 0,
    "det_type": "eiger",
}


chipmap_str = """01status    P3011       1
02status    P3021       0
03status    P3031       0
04status    P3041       0"""


@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.caput")
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.caget")
@patch(
    "mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.scrape_parameter_file"
)
def test_datasetsizei24_for_one_block_and_two_exposures(
    fake_params, fake_caget, fake_caput
):
    fake_params.return_value = tuple(params.values())
    fake_caget.return_value = 2
    with patch(
        "mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.open",
        mock_open(read_data=chipmap_str),
    ):
        tot_num_imgs = datasetsizei24()
    assert tot_num_imgs == 800


def test_get_chip_prog_values():
    chip_dict = get_chip_prog_values(
        0,
        0,
        0,
        0,
        0,
        n_exposures=2,
    )
    assert isinstance(chip_dict, dict)
    assert chip_dict["X_NUM_STEPS"][1] == 20 and chip_dict["X_NUM_BLOCKS"][1] == 8
    assert chip_dict["PUMP_REPEAT"][1] == 0
    assert chip_dict["N_EXPOSURES"][1] == 2


@pytest.mark.parametrize(
    "chip_type, map_type, pump_repeat, expected_prog",
    [
        (0, 0, 0, 11),  # Oxford chip, full chip, no pump
        (0, 1, 0, 12),  # Oxford chip, map generated by mapping lite, no pump
        (2, "", 0, 11),  # Custom chip, map type not needed(full assumed), no pump
        (0, "", 2, 14),  # Oxford chip, assumes mapping lite, pump 2
        (3, "", 0, 11),  # Minichip, no map type, no pump probe
    ],
)
def test_get_prog_number(chip_type, map_type, pump_repeat, expected_prog):
    assert get_prog_num(chip_type, map_type, pump_repeat) == expected_prog


@pytest.mark.parametrize(
    "map_type, pump_repeat, checker, expected_calls",
    [
        (0, 0, False, ["P1100=1"]),  # Full chip, no pump probe, no checker
        (1, 0, False, ["P1200=1"]),  # Mapping lite, no pp, no checker
        (
            1,
            2,
            False,
            ["P1439=0", "P1400=1"],
        ),  # Map irrelevant, pp to Repeat1, no checker
        (
            0,
            3,
            True,
            ["P1439=0", "P1439=1", "P1400=1"],
        ),  # Map irrelevant, pp to Repeat2, checker enabled
    ],
)
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.i24.pmac")
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.caget")
def test_load_motion_program_data_(
    fake_caget, fake_pmac, map_type, pump_repeat, checker, expected_calls
):
    test_dict = {"N_EXPOSURES": [0, 1]}
    fake_caget.return_value = checker
    fake_pmac.pmac_string = MagicMock()
    load_motion_program_data(fake_pmac, test_dict, map_type, pump_repeat)
    call_list = [call.set(i) for i in expected_calls]
    fake_pmac.pmac_string.assert_has_calls(call_list)


@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.datasetsizei24")
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.DCID")
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.caput")
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.caget")
@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.sup")
@patch(
    "mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_Collect_py3v1.scrape_parameter_file"
)
def test_start_i24_with_eiger(
    fake_params, fake_sup, fake_caget, fake_caput, fake_dcid, fake_size
):
    fake_size.return_value = 800
    fake_params.return_value = tuple(params.values())
    start_i24()
    assert fake_sup.beamline.call_count == 2
    assert fake_sup.eiger.call_count == 1
    # Pilatus gets called for hack to create directory
    assert fake_sup.pilatus.call_count == 2
    assert fake_dcid.call_count == 1



import os
import pathlib
import pprint
import time

import requests

from mx_bluesky.I24.serial.fixed_target.ft_utils import ChipType, MappingType
from mx_bluesky.I24.serial.setup_beamline import Eiger, caget, cagetstring, pv
from mx_bluesky.I24.serial.setup_beamline.setup_detector import get_detector_type


def call_nexgen(
    chip_prog_dict,
    start_time,
    param_file_tuple,
    expt_type="fixed-target",
    total_numb_imgs=None,
):
    det_type = get_detector_type()
    print(f"det_type: {det_type}")

    if expt_type == "fixed-target":
        (
            chip_name,
            visit,
            sub_dir,
            n_exposures,
            chip_type,
            map_type,
            pump_repeat,
            pumpexptime,
            pumpdelay,
            exptime,
            dcdetdist,
            prepumpexptime,
            det_type,
        ) = param_file_tuple
        if map_type == MappingType.NoMap or chip_type == ChipType.Custom:
            currentchipmap = "fullchip"
        else:
            currentchipmap = "/dls_sw/i24/scripts/fastchips/litemaps/currentchip.map"
    elif expt_type == "extruder":
        # chip_prog_dict should be None for extruder (passed as input for now)
        (
            visit,
            sub_dir,
            filename,
            num_imgs,
            exptime,
            dcdetdist,
            det_type,
            pump_status,
            pumpexptime,
            pumpdelay,
        ) = param_file_tuple
        total_numb_imgs = num_imgs
        currentchipmap = None
        pump_repeat = "0" if pump_status == "false" else "1"

    filename_prefix = cagetstring(pv.eiger_ODfilenameRBV)
    meta_h5 = pathlib.Path(visit) / sub_dir / f"{filename_prefix}_meta.h5"
    t0 = time.time()
    max_wait = 60  # seconds
    print(f"Watching for {meta_h5}")
    while time.time() - t0 < max_wait:
        if meta_h5.exists():
            print(f"Found {meta_h5} after {time.time() - t0:.1f} seconds")
            time.sleep(5)
            break
        print(f"Waiting for {meta_h5}")
        time.sleep(1)
    if not meta_h5.exists():
        print(f"Giving up waiting for {meta_h5} after {max_wait} seconds")
        return False

    # filepath = visit + sub_dir
    # filename = chip_name
    transmission = (float(caget(pv.pilat_filtertrasm)),)
    wavelength = float(caget(pv.dcm_lambda))

    if det_type == Eiger.name:
        print("nexgen here")
        print(chip_prog_dict)

        access_token = pathlib.Path("/scratch/ssx_nexgen.key").read_text().strip()
        url = "https://ssx-nexgen.diamond.ac.uk/ssx_eiger/write"
        headers = {"Authorization": f"Bearer {access_token}"}

        payload = {
            "beamline": "i24",
            "beam_center": [caget(pv.eiger_beamx), caget(pv.eiger_beamy)],
            "chipmap": currentchipmap,
            "chip_info": chip_prog_dict,
            "det_dist": dcdetdist,
            "exp_time": exptime,
            "expt_type": expt_type,
            "filename": filename_prefix,
            "num_imgs": int(total_numb_imgs),
            "pump_status": bool(float(pump_repeat)),
            "pump_exp": pumpexptime,
            "pump_delay": pumpdelay,
            "transmission": transmission[0],
            "visitpath": os.fspath(meta_h5.parent),
            "wavelength": wavelength,
        }
        print(f"Sending POST request to {url} with payload:")
        pprint.pprint(payload)
        response = requests.post(url, headers=headers, json=payload)
        print(f"Response: {response.text} (status code: {response.status_code})")
        # the following will raise an error if the request was unsuccessful
        return response.status_code == requests.codes.ok
    return False



from unittest.mock import patch

from bluesky import RunEngine
from ophyd.sim import make_fake_device

from mx_bluesky.example import Synchrotron, test_plan


@patch("mx_bluesky.example.print")
def test_example_reads_correct_value(mock_print):
    fake_device: Synchrotron = make_fake_device(Synchrotron)(name="fake_synch")
    fake_device.ring_current.sim_put(378.8)
    RE = RunEngine()
    RE(test_plan(fake_device))

    assert mock_print.called_once_with(str(378.8))



import argparse
from unittest.mock import mock_open, patch

import pytest

from mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2 import (
    initialise_extruderi24,
    moveto,
    run_extruderi24,
    scrape_parameter_file,
)
from mx_bluesky.I24.serial.setup_beamline import Eiger, Pilatus

params_file_str = """visit foo
directory bar
filename boh
num_imgs 1
nexp_time 0.1
det_dist 100
det_type eiger
pump_probe false
pump_exp 0
pump_delay 0"""


@pytest.fixture
def dummy_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "place",
        type=str,
        choices=["laseron", "laseroff", "enterhutch"],
        help="Requested setting.",
    )
    yield parser


@patch(
    "mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2.open",
    mock_open(read_data=params_file_str),
)
def test_scrape_parameter_file():
    res = scrape_parameter_file()
    assert res[0] == "foo"
    assert len(res) == 10


@patch("mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2.caget")
@patch("mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2.caput")
@patch("mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2.get_detector_type")
@patch("mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2.logger")
def test_initialise_extruder(fake_log, fake_det, fake_caput, fake_caget):
    fake_caget.return_value = "/path/to/visit"
    fake_det.return_value = Eiger()
    initialise_extruderi24()
    assert fake_caput.call_count == 10
    assert fake_caget.call_count == 1


@patch("mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2.caput")
@patch("mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2.get_detector_type")
def test_moveto_enterhutch(fake_det, fake_caput, dummy_parser):
    fake_args = dummy_parser.parse_args(["enterhutch"])
    fake_det.return_value = Eiger()
    moveto(fake_args)
    assert fake_caput.call_count == 1


@patch("mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2.caput")
@patch("mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2.get_detector_type")
def test_moveto_laseron_for_eiger(fake_det, fake_caput, dummy_parser):
    fake_det.return_value = Eiger()
    fake_args = dummy_parser.parse_args(["laseron"])
    moveto(fake_args)
    assert fake_caput.call_count == 2


@patch("mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2.caput")
@patch("mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2.get_detector_type")
def test_moveto_laseroff_for_pilatus(fake_det, fake_caput, dummy_parser):
    fake_det.return_value = Pilatus()
    fake_args = dummy_parser.parse_args(["laseroff"])
    moveto(fake_args)
    assert fake_caput.call_count == 2


@patch(
    "mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2.open",
    mock_open(read_data=params_file_str),
)
@patch("mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2.DCID")
@patch("mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2.call_nexgen")
@patch("mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2.caput")
@patch("mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2.caget")
@patch("mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2.sup")
@patch("mx_bluesky.I24.serial.extruder.i24ssx_Extruder_Collect_py3v2.get_detector_type")
def test_run_extruder_with_eiger(
    fake_det, fake_sup, fake_caget, fake_caput, fake_nexgen, fake_dcid
):
    fake_det.return_value = Eiger()
    run_extruderi24()
    assert fake_nexgen.call_count == 1
    assert fake_dcid.call_count == 1
    # Check temporary piilatus hack is in there
    assert fake_sup.pilatus.call_count == 2



"""
Startup utilities for chip

This version changed to python3 March2020 by RLO
"""
from __future__ import annotations

import logging
import os
import string
import time
from pathlib import Path
from typing import Dict, List

import numpy as np

from mx_bluesky.I24.serial import log
from mx_bluesky.I24.serial.fixed_target.ft_utils import ChipType
from mx_bluesky.I24.serial.parameters.constants import (
    HEADER_FILES_PATH,
    PARAM_FILE_PATH_FT,
)

logger = logging.getLogger("I24ssx.chip_startup")


def setup_logging():
    # Log should now change name daily.
    logfile = time.strftime("i24fixedtarget_%d%B%y.log").lower()
    log.config(logfile)


def scrape_parameter_file(param_path: Path | str = PARAM_FILE_PATH_FT):
    if not isinstance(param_path, Path):
        param_path = Path(param_path)

    with open(param_path / "parameters.txt", "r") as filein:
        f = filein.readlines()
    for line in f:
        entry = line.rstrip().split()
        if "chip_name" in entry[0].lower():
            chip_name = entry[1]
        elif line.startswith("visit"):
            visit = entry[1]
        elif line.startswith("sub_dir"):
            sub_dir = entry[1]
        elif line.startswith("protein_name"):
            sub_dir = entry[1]
        elif "n_exposures" in entry[0].lower():
            n_exposures = int(entry[1])
        elif "chip_type" in entry[0].lower():
            chip_type = int(entry[1])
        elif "map_type" in entry[0].lower():
            map_type = int(entry[1])
        elif "pump_repeat" in entry[0].lower():
            pump_repeat = int(entry[1])

    for line in f:
        entry = line.rstrip().split()
        if "pumpexptime" == entry[0].lower().strip():
            pumpexptime = float(entry[1])
        if "exptime" in entry[0].lower():
            exptime = float(entry[1])
        if "dcdetdist" in entry[0].lower():
            dcdetdist = float(entry[1])
        if "prepumpexptime" in entry[0].lower():
            prepumpexptime = float(entry[1])
        if "pumpdelay" in entry[0].lower():
            pumpdelay = float(entry[1])
        if "det_type" in entry[0].lower():
            det_type = entry[1]
    return (
        chip_name,
        visit,
        sub_dir,
        n_exposures,
        chip_type,
        map_type,
        pump_repeat,
        pumpexptime,
        pumpdelay,
        exptime,
        dcdetdist,
        prepumpexptime,
        det_type,
    )


def read_parameters(
    param_path: Path | str = PARAM_FILE_PATH_FT, filename: str | None = None
) -> Dict[str, str]:
    """
    Read the parameter file into a lookup dictionary.

    Does the same thing as scrape_parameter_file except doesn't rely on you
    getting the order of arguments right every time (or having to load every one
    if you don't need them all).

    Args:
        filename: The file to read. If None, will load from default location.

    Returns:
        A dictionary with a string entry for every key in the file.
    """
    if not isinstance(param_path, Path):
        param_path = Path(param_path)
    if filename is None:
        filename = "parameters.txt"
    datafile = param_path / filename
    data = datafile.read_text()
    args = {}
    for line in data.splitlines():
        key, value = line.split(maxsplit=1)
        args[key.lower()] = value
    return args


@log.log_on_entry
def fiducials(chip_type: int):
    if chip_type in [ChipType.Oxford, ChipType.OxfordInner, ChipType.Minichip]:
        fiducial_list: list = []
        # No fiducial for custom
    elif chip_type == ChipType.Custom:
        logger.warning("No fiducials for custom chip")
    else:
        logger.warning("Unknown chip_type, %s, in fiducials" % chip_type)
    return fiducial_list


@log.log_on_entry
def get_format(chip_type: int):
    if chip_type == ChipType.Oxford:
        w2w = 0.125
        b2b_horz = 0.800
        b2b_vert = 0.800
        chip_format = [8, 8, 20, 20]
    elif chip_type == ChipType.OxfordInner:
        w2w = 0.600
        b2b_horz = 0.0
        b2b_vert = 0.0
        chip_format = [1, 1, 25, 25]
    elif chip_type == ChipType.Minichip:
        w2w = 0.125
        b2b_horz = 0
        b2b_vert = 0
        chip_format = [1, 1, 20, 20]
    else:
        msg = "Unknown chip_type, %s" % chip_type
        logger.error(msg)
        raise ValueError(msg)
    cell_format = chip_format + [w2w, b2b_horz, b2b_vert]
    logger.info("Cell format for chip type %s: %s" % (chip_type, cell_format))
    return cell_format


def get_xy(addr: str, chip_type: int):
    entry = addr.split("_")[-2:]
    R, C = entry[0][0], entry[0][1]
    r2, c2 = entry[1][0], entry[1][1]
    blockR = string.ascii_uppercase.index(R)
    blockC = int(C) - 1
    lowercase_list = list(string.ascii_lowercase + string.ascii_uppercase + "0")
    windowR = lowercase_list.index(r2)
    windowC = lowercase_list.index(c2)

    (
        x_block_num,
        y_block_num,
        x_window_num,
        y_window_num,
        w2w,
        b2b_horz,
        b2b_vert,
    ) = get_format(chip_type)

    x = (blockC * b2b_horz) + (blockC * (x_window_num - 1) * w2w) + (windowC * w2w)
    y = (blockR * b2b_vert) + (blockR * (y_window_num - 1) * w2w) + (windowR * w2w)
    return x, y


def pathli(l_in=[], way="typewriter", reverse=False):
    if reverse is True:
        li = list(reversed(l_in))
    else:
        li = list(l_in)
    long_list = []
    if li:
        if way == "typewriter":
            for i in range(len(li) ** 2):
                long_list.append(li[i % len(li)])
        elif way == "snake":
            lr = list(reversed(li))
            for rep in range(len(li)):
                if rep % 2 == 0:
                    long_list += li
                else:
                    long_list += lr
        elif way == "snake53":
            lr = list(reversed(li))
            for rep in range(53):
                if rep % 2 == 0:
                    long_list += li
                else:
                    long_list += lr
        elif way == "expand":
            for entry in li:
                for rep in range(len(li)):
                    long_list.append(entry)
        elif way == "expand28":
            for entry in li:
                for rep in range(28):
                    long_list.append(entry)
        elif way == "expand25":
            for entry in li:
                for rep in range(25):
                    long_list.append(entry)
        else:
            logger.warning("No known path, way =  %s" % way)
    else:
        logger.warning("No list written")
    return long_list


def zippum(list_1_args, list_2_args):
    list_1, type_1, reverse_1 = list_1_args
    list_2, type_2, reverse_2 = list_2_args
    A_path = pathli(list_1, type_1, reverse_1)
    B_path = pathli(list_2, type_2, reverse_2)
    zipped_list = []
    for a, b in zip(A_path, B_path):
        zipped_list.append(a + b)
    return zipped_list


def get_alphanumeric(chip_type):
    cell_format = get_format(chip_type)
    blk_num = cell_format[0]
    wnd_num = cell_format[2]
    uppercase_list = list(string.ascii_uppercase)[:blk_num]
    lowercase_list = list(string.ascii_lowercase + string.ascii_uppercase + "0")[
        :wnd_num
    ]
    number_list = [str(x) for x in range(1, blk_num + 1)]

    block_list = zippum([uppercase_list, "expand", 0], [number_list, "typewriter", 0])
    window_list = zippum(
        [lowercase_list, "expand", 0], [lowercase_list, "typewriter", 0]
    )

    alphanumeric_list = []
    for block in block_list:
        for window in window_list:
            alphanumeric_list.append(block + "_" + window)
    logger.info("Length of alphanumeric list = %s" % len(alphanumeric_list))
    return alphanumeric_list


@log.log_on_entry
def get_shot_order(chip_type):
    cell_format = get_format(chip_type)
    blk_num = cell_format[0]
    wnd_num = cell_format[2]
    uppercase_list = list(string.ascii_uppercase)[:blk_num]
    number_list = [str(x) for x in range(1, blk_num + 1)]
    lowercase_list = list(string.ascii_lowercase + string.ascii_uppercase + "0")[
        :wnd_num
    ]

    block_list = zippum([uppercase_list, "snake", 0], [number_list, "expand", 0])
    window_dn = zippum([lowercase_list, "expand", 0], [lowercase_list, "snake", 0])
    window_up = zippum([lowercase_list, "expand", 1], [lowercase_list, "snake", 0])

    switch = 0
    count = 0
    collect_list = []
    for block in block_list:
        if switch == 0:
            for window in window_dn:
                collect_list.append(block + "_" + window)
            count += 1
            if count == blk_num:
                count = 0
                switch = 1
        else:
            for window in window_up:
                collect_list.append(block + "_" + window)
            count += 1
            if count == blk_num:
                count = 0
                switch = 0

    logger.info("Length of collect list = %s" % len(collect_list))
    return collect_list


@log.log_on_entry
def write_file(
    location: str = "i24",
    suffix: str = ".addr",
    order: str = "alphanumeric",
    param_file_path: Path = PARAM_FILE_PATH_FT,
    save_path: Path = HEADER_FILES_PATH,
):
    if location == "i24":
        (
            chip_name,
            visit,
            sub_dir,
            n_exposures,
            chip_type,
            map_type,
            pump_repeat,
            pumpexptime,
            exptime,
            dcdetdist,
            prepumpexptime,
        ) = scrape_parameter_file(param_file_path)
    else:
        msg = "Unknown location, %s" % location
        logger.error(msg)
        raise ValueError(msg)
    chip_file_path = save_path / f"chips/{sub_dir}/{chip_name}{suffix}"

    fiducial_list = fiducials(chip_type)
    if order == "alphanumeric":
        addr_list = get_alphanumeric(chip_type)

    elif order == "shot":
        addr_list = get_shot_order(chip_type)

    with open(chip_file_path, "a") as g:
        for addr in addr_list:
            xtal_name = "_".join([chip_name, addr])
            (x, y) = get_xy(xtal_name, chip_type)
            if addr in fiducial_list:
                pres = "0"
            else:
                if "rand" in suffix:
                    pres = str(np.random.randint(2))
                else:
                    pres = "-1"
            line = "\t".join([xtal_name, str(x), str(y), "0.0", pres]) + "\n"
            g.write(line)

    logger.info("Write %s completed" % chip_file_path)


@log.log_on_entry
def check_files(
    location: str,
    suffix_list: List[str],
    param_file_path: Path | str = PARAM_FILE_PATH_FT,
    save_path: Path = HEADER_FILES_PATH,
):
    if location == "i24":
        (
            chip_name,
            visit,
            sub_dir,
            n_exposures,
            chip_type,
            map_type,
            exptime,
            pump_repeat,
            pumpdelay,
            pumpexptime,
            dcdetdist,
            prepumpexptime,
            det_type,
        ) = scrape_parameter_file(param_path=param_file_path)
    else:
        msg = "Unknown location, %s" % location
        logger.error(msg)
        raise ValueError(msg)
    chip_file_path = save_path / f"chips/{sub_dir}/{chip_name}"

    try:
        os.stat(chip_file_path)
    except Exception:
        os.makedirs(chip_file_path)
    for suffix in suffix_list:
        full_fid = chip_file_path.with_suffix(suffix)
        if full_fid.is_file():
            time_str = time.strftime("%Y%m%d_%H%M%S_")
            timestamp_fid = (  # noqa: F841
                full_fid.parent / f"{time_str}_{chip_name}{full_fid.suffix}"
            )
            # FIXME hack / fix. Actually move the file
            logger.info("File %s Already Exists" % full_fid)
    logger.debug("Check files done")
    return 1


@log.log_on_entry
def write_headers(
    location: str,
    suffix_list: List[str],
    param_file_path: Path = PARAM_FILE_PATH_FT,
    save_path: Path = HEADER_FILES_PATH,
):
    if location == "i24":
        (
            chip_name,
            visit,
            sub_dir,
            n_exposures,
            chip_type,
            map_type,
            pump_repeat,
            pumpexptime,
            pumpdelay,
            exptime,
            dcdetdist,
            prepumpexptime,
            det_type,
        ) = scrape_parameter_file(param_path=PARAM_FILE_PATH_FT)
        chip_file_path = save_path / f"chips/{sub_dir}/{chip_name}"

        for suffix in suffix_list:
            full_fid = chip_file_path.with_suffix(suffix)
            with open(full_fid, "w") as g:
                g.write(
                    "#23456789012345678901234567890123456789012345678901234567890123456789012345678901234567890\n#\n"
                )
                g.write("#&i24\tchip_name    = %s\n" % chip_name)
                g.write("#&i24\tvisit        = %s\n" % visit)
                g.write("#&i24\tsub_dir      = %s\n" % sub_dir)
                g.write("#&i24\tn_exposures  = %s\n" % n_exposures)
                g.write("#&i24\tchip_type    = %s\n" % chip_type)
                g.write("#&i24\tmap_type     = %s\n" % map_type)
                g.write("#&i24\tpump_repeat  = %s\n" % pump_repeat)
                g.write("#&i24\tpumpexptime  = %s\n" % pumpexptime)
                g.write("#&i24\texptime      = %s\n" % exptime)
                g.write("#&i24\tdcdetdist    = %s\n" % dcdetdist)
                g.write("#&i24\tprepumpexptime  = %s\n" % prepumpexptime)
                g.write("#&i24\tdet_Type     = %s\n" % det_type)
                g.write("#\n")
                g.write(
                    "#XtalAddr      XCoord  YCoord  ZCoord  Present Shot  Spare04 Spare03 Spare02 Spare01\n"
                )
    else:
        msg = "Unknown location, %s" % location
        logger.error(msg)
        raise ValueError(msg)
    logger.debug("Write headers done")


def run():
    logger.debug("Run Startup")
    check_files("i24", [".addr", ".shot"])
    logger.info("Checked Files")
    write_headers("i24", [".addr", ".shot"])
    logger.info("Written Headers")
    logger.info("Writing to Files has been disabled. Headers Only")
    # Makes a file with random crystal positions
    check_files("i24", ["rando.spec"])
    write_headers("i24", ["rando.spec"])
    logger.debug("StartUp Done")


if __name__ == "__main__":
    setup_logging()
    run()



"""
Extruder data collection
This version in python3 new Feb2021 by RLO
    - March 21 added logging and Eiger functionality
"""
from __future__ import annotations

import argparse
import logging
import re
import sys
import time
from datetime import datetime
from pathlib import Path
from time import sleep

from mx_bluesky.I24.serial import log
from mx_bluesky.I24.serial.dcid import DCID
from mx_bluesky.I24.serial.parameters import SSXType
from mx_bluesky.I24.serial.parameters.constants import PARAM_FILE_PATH
from mx_bluesky.I24.serial.setup_beamline import Eiger, Pilatus, caget, caput, pv
from mx_bluesky.I24.serial.setup_beamline import setup_beamline as sup
from mx_bluesky.I24.serial.setup_beamline.setup_detector import get_detector_type
from mx_bluesky.I24.serial.write_nexus import call_nexgen

usage = "%(prog)s command [options]"
logger = logging.getLogger("I24ssx.extruder")


def setup_logging():
    logfile = time.strftime("i24extruder_%d%B%y.log").lower()
    log.config(logfile)


def flush_print(text):
    sys.stdout.write(str(text))
    sys.stdout.flush()


def _coerce_to_path(path: Path | str) -> Path:
    if not isinstance(path, Path):
        return Path(path)
    return path


@log.log_on_entry
def initialise_extruderi24(args=None):
    logger.info("Initialise Parameters for extruder data collection on I24.")

    visit = caget(pv.ioc12_gp1)
    logger.info("Visit defined %s" % visit)

    # Define detector in use
    det_type = get_detector_type()

    caput(pv.ioc12_gp2, "test")
    caput(pv.ioc12_gp3, "testrun")
    caput(pv.ioc12_gp4, "100")
    caput(pv.ioc12_gp5, "0.01")
    caput(pv.ioc12_gp6, 0)
    caput(pv.ioc12_gp8, 0)  # status PV do not reuse gp8 for something else
    caput(pv.ioc12_gp9, 0)
    caput(pv.ioc12_gp10, 0)
    caput(pv.ioc12_gp15, det_type.name)
    caput(pv.pilat_cbftemplate, 0)
    logger.info("Initialisation complete.")


@log.log_on_entry
def moveto(args):
    place = args.place
    logger.info("Move to: %s" % place)

    det_type = get_detector_type()

    if place == "laseron":
        if isinstance(det_type, Pilatus):
            caput(pv.zebra1_out1_ttl, 60.0)
            caput(pv.zebra1_soft_in_b0, 1.0)
        elif isinstance(det_type, Eiger):
            caput(pv.zebra1_out2_ttl, 60.0)
            caput(pv.zebra1_soft_in_b0, 1.0)

    if place == "laseroff":
        if isinstance(det_type, Pilatus):
            caput(pv.zebra1_soft_in_b0, 0.0)
            caput(pv.zebra1_out1_ttl, 0.0)
        elif isinstance(det_type, Eiger):
            caput(pv.zebra1_soft_in_b0, 0.0)
            caput(pv.zebra1_out2_ttl, 0.0)

    if place == "enterhutch":
        caput(pv.det_z, 1480)


@log.log_on_entry
def write_parameter_file(param_path: Path | str = PARAM_FILE_PATH):
    param_path = _coerce_to_path(param_path)
    param_fid = "parameters.txt"

    logger.info("Writing Parameter File to: %s \n" % (param_path / param_fid))

    visit = caget(pv.ioc12_gp1)
    directory = caget(pv.ioc12_gp2)
    filename = caget(pv.ioc12_gp3)
    num_imgs = caget(pv.ioc12_gp4)
    exp_time = caget(pv.ioc12_gp5)
    det_dist = caget(pv.ioc12_gp7)
    det_type = get_detector_type()
    if int(caget(pv.ioc12_gp6)) == 1:
        pump_status = "true"
    else:
        pump_status = "false"
    pump_exp = caget(pv.ioc12_gp9)
    pump_delay = caget(pv.ioc12_gp10)

    # If file name ends in a digit this causes processing/pilatus pain.
    # Append an underscore
    if det_type.name == "pilatus":
        m = re.search(r"\d+$", filename)
        if m is not None:
            # Note for future reference. Appending underscore causes more hassle and
            # high probability of users accidentally overwriting data. Use a dash
            filename = filename + "-"
            logger.info(
                "Requested filename ends in a number. Appended dash: %s" % filename
            )

    with open(param_path / param_fid, "w") as f:
        f.write("visit \t\t%s\n" % visit)
        f.write("directory \t%s\n" % directory)
        f.write("filename \t%s\n" % filename)
        f.write("num_imgs \t%s\n" % num_imgs)
        f.write("exp_time \t%s\n" % exp_time)
        f.write("det_dist \t%s\n" % det_dist)
        f.write("det_type \t%s\n" % det_type.name)
        f.write("pump_probe \t%s\n" % pump_status)
        f.write("pump_exp \t%s\n" % pump_exp)
        f.write("pump_delay \t%s\n" % pump_delay)

    logger.info("Parameters \n")
    logger.info("visit %s" % visit)
    logger.info("directory %s" % directory)
    logger.info("filename %s" % filename)
    logger.info("num_imgs %s" % num_imgs)
    logger.info("exp_time %s" % exp_time)
    logger.info("det_dist %s" % det_dist)
    logger.info("det_type %s" % det_type.name)
    logger.info("pump_probe %s" % pump_status)
    logger.info("pump_exp %s" % pump_exp)
    logger.info("pump_delay %s" % pump_delay)


def scrape_parameter_file(param_path: Path | str = PARAM_FILE_PATH):
    param_path = _coerce_to_path(param_path)

    with open(param_path / "parameters.txt", "r") as filein:
        f = filein.readlines()
    for line in f:
        entry = line.rstrip().split()
        if line.startswith("visit"):
            visit = entry[1]
        elif line.startswith("directory"):
            directory = entry[1]
        elif line.startswith("filename"):
            filename = entry[1]
        elif "num_imgs" in entry[0].lower():
            num_imgs = entry[1]
        elif "exp_time" in entry[0].lower():
            exp_time = entry[1]
        elif "det_dist" in entry[0].lower():
            det_dist = entry[1]
        elif "det_type" in entry[0].lower():
            det_type = entry[1]
        elif "pump_probe" in entry[0].lower():
            pump_status = entry[1]
        elif "pump_exp" in entry[0].lower():
            pump_exp = entry[1]
        elif "pump_delay" in entry[0].lower():
            pump_delay = entry[1]
    return (
        visit,
        directory,
        filename,
        num_imgs,
        exp_time,
        det_dist,
        det_type,
        pump_status,
        pump_exp,
        pump_delay,
    )


@log.log_on_entry
def run_extruderi24(args=None):
    start_time = datetime.now()
    logger.info("Collection start time: %s" % start_time.ctime())

    write_parameter_file()
    (
        visit,
        directory,
        filename,
        num_imgs,
        exp_time,
        det_dist,
        det_type,
        pump_status,
        pump_exp,
        pump_delay,
    ) = scrape_parameter_file()

    # Setting up the beamline
    caput("BL24I-PS-SHTR-01:CON", "Reset")
    logger.debug("Reset hutch shutter sleep for 1sec")
    sleep(1.0)
    caput("BL24I-PS-SHTR-01:CON", "Open")
    logger.debug("Open hutch shutter sleep for 2sec")
    sleep(2.0)

    sup.beamline("collect")
    sup.beamline("quickshot", [det_dist])

    # Set the abort PV to zero
    caput(pv.ioc12_gp8, 0)

    # For pixel detector
    filepath = visit + directory
    logger.info("Filepath %s" % filepath)
    logger.info("Filename %s" % filename)

    # For zebra
    # The below will need to be determined emprically. A value of 0.0 may be ok (????)
    probepumpbuffer = 0.01

    gate_start = 1.0
    # Need to check these for pilatus.
    # Added temprary hack in pilatus pump is false below as gate width wrong
    gate_width = float(pump_exp) + float(pump_delay) + float(exp_time)
    gate_step = float(gate_width) + float(probepumpbuffer)
    logger.info("Calculated gate width %.4f" % gate_width)
    logger.info("Calculated gate step %.4f" % gate_step)
    num_gates = num_imgs
    p1_delay = 0
    p1_width = pump_exp
    p2_delay = pump_delay
    p2_width = exp_time

    if det_type == "pilatus":
        logger.debug("Using pilatus mini cbf")
        caput(pv.pilat_cbftemplate, 0)
        logger.info("Pilatus quickshot setup: filepath %s" % filepath)
        logger.info("Pilatus quickshot setup: filepath %s" % filename)
        logger.info("Pilatus quickshot setup: number of images %d" % num_imgs)
        logger.info("Pilatus quickshot setup: exposure time %s" % exp_time)

        if pump_status == "true":
            logger.info("Pump probe extruder data collection")
            logger.info("Pump exposure time %s" % pump_exp)
            logger.info("Pump delay time %s" % pump_delay)
            sup.pilatus("fastchip", [filepath, filename, num_imgs, exp_time])
            sup.zebra1(
                "zebratrigger-pilatus",
                [
                    gate_start,
                    gate_width,
                    num_gates,
                    gate_step,
                    p1_delay,
                    p1_width,
                    p2_delay,
                    p2_width,
                ],
            )
        elif pump_status == "false":
            logger.info("Static experiment: no photoexcitation")
            sup.pilatus("quickshot", [filepath, filename, num_imgs, exp_time])
            gate_start = 1.0
            gate_width = (float(exp_time) * float(num_imgs)) + float(0.5)
            sup.zebra1("quickshot", [gate_start, gate_width])

    elif det_type == "eiger":
        logger.info("Using Eiger detector")

        logger.warning(
            """TEMPORARY HACK!
            Running a Single image pilatus data collection to create directory."""
        )  # See https://github.com/DiamondLightSource/mx_bluesky/issues/45
        num_shots = 1
        sup.pilatus("quickshot-internaltrig", [filepath, filename, num_shots, exp_time])
        logger.debug("Sleep 2s waiting for pilatus to arm")
        sleep(2.5)
        caput(pv.pilat_acquire, "0")  # Disarm pilatus
        sleep(0.5)
        caput(pv.pilat_acquire, "1")  # Arm pilatus
        logger.debug("Pilatus data collection DONE")
        sup.pilatus("return to normal")
        logger.info("Pilatus back to normal. Single image pilatus data collection DONE")

        caput(pv.eiger_seqID, int(caget(pv.eiger_seqID)) + 1)
        logger.info("Eiger quickshot setup: filepath %s" % filepath)
        logger.info("Eiger quickshot setup: filepath %s" % filename)
        logger.info("Eiger quickshot setup: number of images %s" % num_imgs)
        logger.info("Eiger quickshot setup: exposure time %s" % exp_time)

        if pump_status == "true":
            logger.info("Pump probe extruder data collection")
            logger.info("Pump exposure time %s" % pump_exp)
            logger.info("Pump delay time %s" % pump_delay)
            sup.eiger("triggered", [filepath, filename, num_imgs, exp_time])
            sup.zebra1(
                "zebratrigger-eiger",
                [
                    gate_start,
                    gate_width,
                    num_gates,
                    gate_step,
                    p1_delay,
                    p1_width,
                    p2_delay,
                    p2_width,
                ],
            )
        elif pump_status == "false":
            logger.info("Static experiment: no photoexcitation")
            gate_start = 1.0
            gate_width = (float(exp_time) * float(num_imgs)) + float(0.5)
            sup.eiger("quickshot", [filepath, filename, num_imgs, exp_time])
            sup.zebra1("quickshot", [gate_start, gate_width])
    else:
        err = "Unknown Detector Type, det_type = %s" % det_type
        logger.error(err)
        raise ValueError(err)

    # Do DCID creation BEFORE arming the detector
    dcid = DCID(
        emit_errors=False,
        ssx_type=SSXType.EXTRUDER,
        visit=Path(visit).name,
        image_dir=filepath,
        start_time=start_time,
        num_images=num_imgs,
        exposure_time=exp_time,
    )

    # Collect
    logger.info("Fast shutter opening")
    caput(pv.zebra1_soft_in_b1, 1)
    if det_type == "pilatus":
        logger.info("Pilatus acquire ON")
        caput(pv.pilat_acquire, 1)
    elif det_type == "eiger":
        logger.info("Triggering Eiger NOW")
        caput(pv.eiger_trigger, 1)

    dcid.notify_start()

    param_file_tuple = scrape_parameter_file()
    if det_type == "eiger":
        logger.info("Call nexgen server for nexus writing.")
        call_nexgen(None, start_time, param_file_tuple, "extruder")

    aborted = False
    timeout_time = time.time() + int(num_imgs) * float(exp_time) + 10

    if int(caget(pv.ioc12_gp8)) == 0:  # ioc12_gp8 is the ABORT button
        caput(pv.zebra1_pc_arm, 1)
        sleep(gate_start)
        i = 0
        text_list = ["|", "/", "-", "\\"]
        while True:
            line_of_text = "\r\t\t\t Waiting   " + 30 * ("%s" % text_list[i % 4])
            flush_print(line_of_text)
            sleep(0.5)
            i += 1
            if int(caget(pv.ioc12_gp8)) != 0:
                aborted = True
                logger.warning("Data Collection Aborted")
                if det_type == "pilatus":
                    caput(pv.pilat_acquire, 0)
                elif det_type == "eiger":
                    caput(pv.eiger_acquire, 0)
                sleep(1.0)
                break
            elif int(caget(pv.zebra1_pc_arm_out)) != 1:
                # As soon as the zebra1_pc_arm_out is not 1 anymore, exit.
                # Epics checks the geobrick and updates this PV once the collection is done.
                logger.info("----> Zebra disarmed  <----")
                break
            elif time.time() >= timeout_time:
                logger.warning(
                    """
                    Something went wrong and data collection timed out. Aborting.
                """
                )
                if det_type == "pilatus":
                    caput(pv.pilat_acquire, 0)
                elif det_type == "eiger":
                    caput(pv.eiger_acquire, 0)
                sleep(1.0)
                break
    else:
        aborted = True
        logger.warning("Data Collection ended due to GP 8 not equalling 0")

    caput(pv.ioc12_gp8, 1)
    logger.info("Fast shutter closing")
    caput(pv.zebra1_soft_in_b1, 0)
    logger.info("\nZebra DISARMED")
    caput(pv.zebra1_pc_disarm, 1)

    end_time = datetime.now()

    if det_type == "pilatus":
        logger.info("Pilatus Acquire STOP")
        caput(pv.pilat_acquire, 0)
    elif det_type == "eiger":
        logger.info("Eiger Acquire STOP")
        caput(pv.eiger_acquire, 0)
        caput(pv.eiger_ODcapture, "Done")

    sleep(0.5)

    # Clean Up
    logger.info("Setting zebra back to normal")
    sup.zebra1("return-to-normal")
    if det_type == "pilatus":
        sup.pilatus("return-to-normal")
    elif det_type == "eiger":
        sup.eiger("return-to-normal")
        print(filename + "_" + caget(pv.eiger_seqID))
    logger.info("End of Run")
    logger.info("Close hutch shutter")
    caput("BL24I-PS-SHTR-01:CON", "Close")

    dcid.collection_complete(end_time, aborted=aborted)
    dcid.notify_end()
    logger.info("End Time = %s" % end_time.ctime())
    return 1


if __name__ == "__main__":
    setup_logging()

    parser = argparse.ArgumentParser(usage=usage, description=__doc__)
    subparsers = parser.add_subparsers(
        help="Choose command.",
        required=True,
        dest="sub-command",
    )

    parser_init = subparsers.add_parser(
        "initialise",
        description="Initialise extruder on beamline I24.",
    )
    parser_init.set_defaults(func=initialise_extruderi24)
    parser_run = subparsers.add_parser(
        "run",
        description="Run extruder on I24.",
    )
    parser_run.set_defaults(func=run_extruderi24)
    parser_mv = subparsers.add_parser(
        "moveto",
        description="Move extruder to requested setting on I24.",
    )
    parser_mv.add_argument(
        "place",
        type=str,
        choices=["laseron", "laseroff", "enterhutch"],
        help="Requested setting.",
    )
    parser_mv.set_defaults(func=moveto)

    args = parser.parse_args()
    args.func(args)



"""
Utilities for defining the detector in use, and moving the stage.
"""
import argparse
import logging
import time
from enum import IntEnum

import bluesky.plan_stubs as bps
from bluesky.run_engine import RunEngine
from dodal.beamlines import i24
from dodal.devices.i24.I24_detector_motion import DetectorMotion

from mx_bluesky.I24.serial import log
from mx_bluesky.I24.serial.parameters.constants import SSXType
from mx_bluesky.I24.serial.setup_beamline import pv
from mx_bluesky.I24.serial.setup_beamline.ca import caget
from mx_bluesky.I24.serial.setup_beamline.pv_abstract import (
    Detector,
    Eiger,
    Pilatus,
)

logger = logging.getLogger("I24ssx.sup_det")

EXPT_TYPE_DETECTOR_PVS = {
    SSXType.FIXED: pv.me14e_gp101,
    SSXType.EXTRUDER: pv.ioc12_gp15,
}


class DetRequest(IntEnum):
    eiger = 0
    pilatus = 1


def setup_logging():
    logfile = time.strftime("SSXdetectorOps_%d%B%y.log").lower()
    log.config(logfile)


class UnknownDetectorType(Exception):
    pass


def get_detector_type() -> Detector:
    det_y = caget(pv.det_y)
    # DetectorMotion should also be used for this.
    # This should be part of https://github.com/DiamondLightSource/mx_bluesky/issues/51
    if float(det_y) < Eiger.det_y_threshold:
        logger.info("Eiger detector in use.")
        return Eiger()
    elif float(det_y) > Pilatus.det_y_threshold:
        logger.info("Pilatus detector in use.")
        return Pilatus()
    else:
        logger.error("Detector not found.")
        raise UnknownDetectorType("Detector not found.")


def _move_detector_stage(detector_stage: DetectorMotion, target: float):
    logger.info(f"Moving detector stage to target position: {target}.")
    yield from bps.abs_set(
        detector_stage.y,
        target,
        wait=True,
    )


def setup_detector_stage(detector_stage: DetectorMotion, expt_type: SSXType):
    # Grab the correct PV depending on experiment
    # Its value is set with MUX on edm screen
    det_type_pv = EXPT_TYPE_DETECTOR_PVS[expt_type]
    det_type = caget(det_type_pv)
    requested_detector = (
        Eiger.name if int(det_type) == DetRequest.eiger else Pilatus.name
    )
    logger.info(f"Requested detector: {requested_detector}.")
    det_y_target = (
        Eiger.det_y_target if requested_detector == "eiger" else Pilatus.det_y_target
    )
    yield from _move_detector_stage(detector_stage, det_y_target)
    logger.info("Detector setup done.")


if __name__ == "__main__":
    setup_logging()
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "expt",
        type=str,
        choices=[expt.value for expt in SSXType],
        help="Type of serial experiment being run.",
    )
    args = parser.parse_args()
    expt_type = SSXType(args.expt)
    RE = RunEngine()
    # Use dodal device for move
    detector_stage = i24.detector_motion()
    RE(setup_detector_stage(detector_stage, expt_type))



from unittest.mock import mock_open, patch

import pytest

from mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_StartUp_py3v1 import (
    check_files,
    fiducials,
    get_format,
    pathli,
    scrape_parameter_file,
)

params_file_str = """visit foo
sub_dir bar
chip_name chip
protein_name protK
n_exposures 1
chip_type 0
map_type 0
dcdetdist 100
det_type eiger
exptime 0.01
pump_repeat 0
pumpexptime 0
prepumpexptime 0
pumpdelay 0"""


@patch(
    "mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_StartUp_py3v1.open",
    mock_open(read_data=params_file_str),
)
def test_scrape_parameter_file():
    res = scrape_parameter_file()
    assert res[0] == "chip"
    assert res[4] == 0 and res[5] == 0  # chip and map type
    assert len(res) == 13


def test_fiducials():
    assert len(fiducials(0)) == 0
    assert len(fiducials(1)) == 0


def test_get_format_for_oxford_chip():
    # oxford chip
    fmt = get_format(0)
    assert fmt == [8, 8, 20, 20, 0.125, 0.800, 0.800]


def test_get_format_for_oxford_minichip():
    # 1 block of oxford chip
    fmt = get_format(3)
    assert fmt == [1, 1, 20, 20, 0.125, 0.0, 0.0]


@patch("mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_StartUp_py3v1.os")
@patch(
    "mx_bluesky.I24.serial.fixed_target.i24ssx_Chip_StartUp_py3v1.open",
    mock_open(read_data=params_file_str),
)
def test_check_files(mock_os):
    check_files("i24", [".a", ".b"])


@pytest.mark.parametrize(
    "list_in, way, reverse, expected_res",
    [
        (
            [1, 2, 3],
            "typewriter",
            False,
            [1, 2, 3] * 3,
        ),  # Result should be list * len(list)
        ([1, 2, 3], "typewriter", True, [3, 2, 1] * 3),  # list[::-1] * len(list)
        ([4, 5], "snake", False, [4, 5, 5, 4]),  # Snakes the list
        ([4, 5], "expand", False, [4, 4, 5, 5]),  # Repeats each value
    ],
)
def test_pathli(list_in, way, reverse, expected_res):
    assert pathli(list_in, way, reverse) == expected_res



import logging
import subprocess
from os import environ
from pathlib import Path

from mx_bluesky.I24.serial import log


def setup_logging():
    logger = logging.getLogger("I24ssx")
    log.config("edm_screen.log")
    return logger


def get_location(default: str = "dev") -> str:
    return environ.get("BEAMLINE") or default


def get_edm_path() -> Path:
    return Path(__file__).parents[4] / "edm_serial"


def run_extruder():
    logger = setup_logging()
    loc = get_location()
    logger.info(f"Running on {loc}.")
    edm_path = get_edm_path()
    logger.info("Starting extruder edm screen...")
    subprocess.run(
        [
            "edm",
            "-x",
            edm_path / "EX-gui/DiamondExtruder-I24-py3v1.edl",
        ]
    )
    logger.info("Edm screen closed.")


def run_fixed_target():
    logger = setup_logging()
    loc = get_location()
    logger.info(f"Running on {loc}.")
    edm_path = get_edm_path()
    logger.info("Starting fixed target edm screen...")
    subprocess.run(
        [
            "edm",
            "-x",
            edm_path / "FT-gui/DiamondChipI24-py3v1.edl",
        ]
    )
    logger.info("Edm screen closed.")



"""
Chip manager for fixed target
This version changed to python3 March2020 by RLO
"""
from __future__ import annotations

import argparse
import json
import logging
import re
import shutil
import sys
import time
from pathlib import Path
from time import sleep

import numpy as np

from mx_bluesky.I24.serial import log
from mx_bluesky.I24.serial.fixed_target import i24ssx_Chip_Mapping_py3v1 as mapping
from mx_bluesky.I24.serial.fixed_target import i24ssx_Chip_StartUp_py3v1 as startup
from mx_bluesky.I24.serial.fixed_target.ft_utils import ChipType, MappingType
from mx_bluesky.I24.serial.parameters.constants import (
    CS_FILES_PATH,
    FULLMAP_PATH,
    LITEMAP_PATH,
    PARAM_FILE_PATH_FT,
    PVAR_FILE_PATH,
)
from mx_bluesky.I24.serial.setup_beamline import caget, caput, pv
from mx_bluesky.I24.serial.setup_beamline.setup_detector import get_detector_type

logger = logging.getLogger("I24ssx.chip_manager")


def _coerce_to_path(path: Path | str) -> Path:
    if not isinstance(path, Path):
        return Path(path)
    return path


def setup_logging():
    # Log should now change name daily.
    logfile = time.strftime("i24fixedtarget_%d%B%y.log").lower()
    log.config(logfile)


@log.log_on_entry
def initialise():
    # commented out filter lines 230719 as this stage not connected
    logger.info("Setting VMAX VELO ACCL HHL LLM pvs")

    caput(pv.me14e_stage_x + ".VMAX", 20)
    caput(pv.me14e_stage_y + ".VMAX", 20)
    caput(pv.me14e_stage_z + ".VMAX", 20)
    # caput(pv.me14e_filter  + '.VMAX', 20)
    caput(pv.me14e_stage_x + ".VELO", 20)
    caput(pv.me14e_stage_y + ".VELO", 20)
    caput(pv.me14e_stage_z + ".VELO", 20)
    # caput(pv.me14e_filter  + '.VELO', 20)
    caput(pv.me14e_stage_x + ".ACCL", 0.01)
    caput(pv.me14e_stage_y + ".ACCL", 0.01)
    caput(pv.me14e_stage_z + ".ACCL", 0.01)
    # caput(pv.me14e_filter  + '.ACCL', 0.01)
    caput(pv.me14e_stage_x + ".HLM", 30)
    caput(pv.me14e_stage_x + ".LLM", -29)
    caput(pv.me14e_stage_y + ".HLM", 30)
    caput(pv.me14e_stage_y + ".LLM", -30)
    # caput(pv.me14e_stage_x + '.LLM', -30)
    caput(pv.me14e_stage_z + ".HLM", 5.1)
    caput(pv.me14e_stage_z + ".LLM", -4.1)
    # caput(pv.me14e_filter  + '.HLM', 45.0)
    # caput(pv.me14e_filter  + '.LLM', -45.0)
    caput(pv.me14e_gp1, 1)
    caput(pv.me14e_gp2, 0)
    caput(pv.me14e_gp3, 1)
    caput(pv.me14e_gp4, 0)
    caput(pv.me14e_filepath, "test")
    caput(pv.me14e_chip_name, "albion")
    caput(pv.me14e_dcdetdist, 1480)
    caput(pv.me14e_exptime, 0.01)
    caput(pv.me14e_pmac_str, "m508=100 m509=150")
    caput(pv.me14e_pmac_str, "m608=100 m609=150")
    caput(pv.me14e_pmac_str, "m708=100 m709=150")
    caput(pv.me14e_pmac_str, "m808=100 m809=150")

    # Define detector in use
    logger.debug("Define detector in use.")
    det_type = get_detector_type()

    caput(pv.pilat_cbftemplate, 0)

    sleep(0.1)
    logger.info("Clearing General Purpose PVs 1-120")
    for i in range(4, 120):
        pvar = "ME14E-MO-IOC-01:GP" + str(i)
        caput(pvar, 0)
        sys.stdout.write(".")
        sys.stdout.flush()

    caput(pv.me14e_gp100, "press set params to read visit")
    caput(pv.me14e_gp101, det_type.name)

    logger.info("Initialisation Complete")


@log.log_on_entry
def write_parameter_file(param_path: Path | str = PARAM_FILE_PATH_FT):
    param_path = _coerce_to_path(param_path)
    param_path.mkdir(parents=True, exist_ok=True)

    param_fid = "parameters.txt"
    logger.info("Writing Parameter File: %s" % (param_path / param_fid).as_posix())

    visit = caget(pv.me14e_gp100)

    filename = caget(pv.me14e_chip_name)

    exptime = caget(pv.me14e_exptime)
    dcdetdist = caget(pv.me14e_dcdetdist)
    protein_name = caget(pv.me14e_filepath)
    pump_repeat = caget(pv.me14e_gp4)
    pumpexptime = caget(pv.me14e_gp103)
    pumpdelay = caget(pv.me14e_gp110)
    prepumpexptime = caget(pv.me14e_gp109)
    n_exposures = caget(pv.me14e_gp3)
    map_type = caget(pv.me14e_gp2)
    chip_type = caget(pv.me14e_gp1)
    det_type = get_detector_type()

    checkerpattern = bool(caget(pv.me14e_gp111))

    # If file name ends in a digit this causes processing/pilatus pain.
    # Append an underscore
    if det_type == "pilatus":
        caput(pv.pilat_cbftemplate, 0)
        m = re.search(r"\d+$", filename)
        if m is not None:
            # Note for future reference. Appending underscore causes more hassle and
            # high probability of users accidentally overwriting data. Use a dash
            filename = filename + "-"
            logger.info(
                "Requested filename ends in a number. Appended dash: %s" % filename
            )

    # historical - use chip_name instead of filename
    chip_name = filename

    with open(param_path / param_fid, "w") as f:
        f.write("visit \t\t%s\n" % visit)
        f.write("chip_name \t%s\n" % chip_name)
        f.write("protein_name \t%s\n" % protein_name)
        f.write("n_exposures \t%s\n" % n_exposures)
        f.write("chip_type \t%s\n" % chip_type)
        f.write("map_type \t%s\n" % map_type)
        f.write("pump_repeat \t%s\n" % pump_repeat)
        f.write("pumpexptime \t%s\n" % pumpexptime)
        f.write("pumpdelay \t%s\n" % pumpdelay)
        f.write("prepumpexptime \t%s\n" % prepumpexptime)
        f.write("exptime \t%s\n" % exptime)
        f.write("dcdetdist \t%s\n" % dcdetdist)
        f.write("det_type \t%s\n" % str(det_type))
        f.write("checkerpattern \t%s\n" % str(checkerpattern))

    logger.info("Information written to file")
    logger.info(f"visit: {visit}")
    logger.info(f"filename: {chip_name}")
    logger.info(f"protein_name: {protein_name}")
    logger.info(f"n_exposures: {n_exposures}")
    logger.info(f"chip_type: {chip_type}")
    logger.info(f"map_type: {map_type}")
    logger.info(f"pump_repeat: {pump_repeat}")
    logger.info(f"pumpexptime: {pumpexptime}")
    logger.info(f"pumpdelay: {pumpdelay}")
    logger.info(f"prepumpexptime: {prepumpexptime}")
    logger.info(f"detector type: {str(det_type)}")
    logger.info(f"checker pattern: {checkerpattern}")
    if map_type == MappingType.Full:
        # This step creates some header files (.addr, .spec), containing the parameters,
        # that are only needed when full mapping is in use.
        logger.debug("Running start up now.")
        startup.run()


def scrape_pvar_file(fid: str, pvar_dir: Path | str = PVAR_FILE_PATH):
    block_start_list = []
    pvar_dir = _coerce_to_path(pvar_dir)

    with open(pvar_dir / fid, "r") as f:
        lines = f.readlines()
    for line in lines:
        line = line.rstrip()
        if line.startswith("#"):
            continue
        elif line.startswith("P3000"):
            continue
        elif line.startswith("P3011"):
            continue
        elif not len(line.split(" ")) == 2:
            continue
        else:
            entry = line.split(" ")
            block_num = entry[0][2:4]
            x = entry[0].split("=")[1]
            y = entry[1].split("=")[1]
            block_start_list.append([block_num, x, y])
    return block_start_list


@log.log_on_entry
def define_current_chip(
    chipid: str = "oxford",
    param_path: Path | str = PVAR_FILE_PATH,
):
    logger.debug("Run load stock map for just the first block")
    load_stock_map("Just The First Block")
    """
    Not sure what this is for:
    print 'Setting Mapping Type to Lite'
    caput(pv.me14e_gp2, 1)
    """
    chip_type = int(caget(pv.me14e_gp1))
    logger.info("Chip type:%s Chipid:%s" % (chip_type, chipid))
    if chipid == "oxford":
        caput(pv.me14e_gp1, 1)

    param_path = _coerce_to_path(param_path)

    with open(param_path / f"{chipid}.pvar", "r") as f:
        logger.info("Opening %s.pvar" % chipid)
        for line in f.readlines():
            if line.startswith("#"):
                continue
            line_from_file = line.rstrip("\n")
            logger.info("%s" % line_from_file)
            caput(pv.me14e_pmac_str, line_from_file)


@log.log_on_entry
def save_screen_map(litemap_path: Path | str = LITEMAP_PATH):
    litemap_path = _coerce_to_path(litemap_path)
    litemap_path.mkdir(parents=True, exist_ok=True)

    logger.info("Saving %s currentchip.map" % litemap_path.as_posix())
    with open(litemap_path / "currentchip.map", "w") as f:
        logger.info("Printing only blocks with block_val == 1")
        for x in range(1, 82):
            block_str = "ME14E-MO-IOC-01:GP%i" % (x + 10)
            block_val = int(caget(block_str))
            if block_val == 1:
                logger.info("%s %d" % (block_str, block_val))
            line = "%02dstatus    P3%02d1 \t%s\n" % (x, x, block_val)
            f.write(line)
    return 0


@log.log_on_entry
def upload_parameters(
    chipid: str = "oxford",
    litemap_path: Path | str = LITEMAP_PATH,
):
    logger.info("Uploading Parameters to the GeoBrick")
    if chipid == "oxford":
        caput(pv.me14e_gp1, 0)
        width = 8
    litemap_path = _coerce_to_path(litemap_path)

    with open(litemap_path / "currentchip.map", "r") as f:
        logger.info("Chipid %s" % chipid)
        logger.info("width %d" % width)
        x = 1
        for line in f.readlines()[: width**2]:
            cols = line.split()
            pvar = cols[1]
            value = cols[2]
            s = pvar + "=" + value
            if value != "1":
                s2 = pvar + "   "
                sys.stdout.write(s2)
            else:
                sys.stdout.write(s + " ")
            sys.stdout.flush()
            if x == width:
                print()
                x = 1
            else:
                x += 1
            caput(pv.me14e_pmac_str, s)
            sleep(0.02)

    logger.warning("Automatic Setting Mapping Type to Lite has been disabled")
    logger.debug("Upload parameters done.")


@log.log_on_entry
def upload_full(fullmap_path: Path | str = FULLMAP_PATH):
    fullmap_path = _coerce_to_path(fullmap_path)

    with open(fullmap_path / "currentchip.full", "r") as fh:
        f = fh.readlines()

    for i in range(len(f) // 2):
        pmac_list = []
        for j in range(2):
            pmac_list.append(f.pop(0).rstrip("\n"))
        writeline = " ".join(pmac_list)
        logger.info("%s" % writeline)
        caput(pv.me14e_pmac_str, writeline)
        sleep(0.02)
    logger.debug("Upload fullmap done")


@log.log_on_entry
def load_stock_map(map_choice: str = "clear"):
    logger.info("Adjusting Lite Map EDM Screen")
    logger.debug("Please wait, adjusting lite map")
    #
    r33 = [19, 18, 17, 26, 31, 32, 33, 24, 25]
    r55 = [9, 10, 11, 12, 13, 16, 27, 30, 41, 40, 39, 38, 37, 34, 23, 20] + r33
    r77 = [
        7,
        6,
        5,
        4,
        3,
        2,
        1,
        14,
        15,
        28,
        29,
        42,
        43,
        44,
        45,
        46,
        47,
        48,
        49,
        36,
        35,
        22,
        21,
        8,
    ] + r55
    #
    h33 = [3, 2, 1, 6, 7, 8, 9, 4, 5]
    x33 = [31, 32, 33, 40, 51, 50, 49, 42, 41]
    x55 = [25, 24, 23, 22, 21, 34, 39, 52, 57, 58, 59, 60, 61, 48, 43, 30] + x33
    x77 = [
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        20,
        35,
        38,
        53,
        56,
        71,
        70,
        69,
        68,
        67,
        66,
        65,
        62,
        47,
        44,
        29,
        26,
    ] + x55
    x99 = [
        9,
        8,
        7,
        6,
        5,
        4,
        3,
        2,
        1,
        18,
        19,
        36,
        37,
        54,
        55,
        72,
        73,
        74,
        75,
        76,
        77,
        78,
        79,
        80,
        81,
        64,
        63,
        46,
        45,
        28,
        27,
        10,
    ] + x77
    x44 = [22, 21, 20, 19, 30, 35, 46, 45, 44, 43, 38, 27, 28, 29, 36, 37]
    x49 = [x + 1 for x in range(49)]
    x66 = [
        10,
        11,
        12,
        13,
        14,
        15,
        18,
        31,
        34,
        47,
        50,
        51,
        52,
        53,
        54,
        55,
        42,
        39,
        26,
        23,
    ] + x44
    x88 = [
        8,
        7,
        6,
        5,
        4,
        3,
        2,
        1,
        16,
        17,
        32,
        33,
        48,
        49,
        64,
        63,
        62,
        61,
        60,
        59,
        58,
        57,
        56,
        41,
        40,
        25,
        24,
        9,
    ] + x66
    #
    # Columns for doing half chips
    c1 = [1, 2, 3, 4, 5, 6, 7, 8]
    c2 = [9, 10, 11, 12, 13, 14, 15, 16]
    c3 = [17, 18, 19, 20, 21, 22, 23, 24]
    c4 = [25, 26, 27, 28, 29, 30, 31, 32]
    c5 = [33, 34, 35, 36, 37, 38, 39, 40]
    c6 = [41, 42, 43, 44, 45, 46, 47, 48]
    c7 = [49, 50, 51, 52, 53, 54, 55, 56]
    c8 = [57, 58, 59, 60, 61, 62, 63, 64]
    half1 = c1 + c2 + c3 + c4
    half2 = c5 + c6 + c7 + c8

    map_dict = {}
    map_dict["Just The First Block"] = [1]
    map_dict["clear"] = []
    #
    map_dict["r33"] = r33
    map_dict["r55"] = r55
    map_dict["r77"] = r77
    #
    map_dict["h33"] = h33
    #
    map_dict["x33"] = x33
    map_dict["x44"] = x44
    map_dict["x49"] = x49
    map_dict["x55"] = x55
    map_dict["x66"] = x66
    map_dict["x77"] = x77
    map_dict["x88"] = x88
    map_dict["x99"] = x99

    map_dict["half1"] = half1
    map_dict["half2"] = half2

    logger.info("Clearing GP 10-74")
    for i in range(1, 65):
        pvar = "ME14E-MO-IOC-01:GP" + str(i + 10)
        caput(pvar, 0)
        sys.stdout.write(".")
        sys.stdout.flush()
    logger.info("Map cleared")
    logger.info("Loading Map Choice %s" % map_choice)
    for i in map_dict[map_choice]:
        pvar = "ME14E-MO-IOC-01:GP" + str(i + 10)
        caput(pvar, 1)
    logger.debug("Load stock map done.")


@log.log_on_entry
def load_lite_map(litemap_path: Path | str = LITEMAP_PATH):
    logger.debug("Run load stock map with 'clear' setting.")
    load_stock_map("clear")
    # fmt: off
    # Oxford_block_dict is wrong (columns and rows need to flip) added in script below to generate it automatically however kept this for backwards compatiability/reference
    oxford_block_dict = {   # noqa: F841
        'A1': '01', 'A2': '02', 'A3': '03', 'A4': '04', 'A5': '05', 'A6': '06', 'A7': '07', 'A8': '08',
        'B1': '16', 'B2': '15', 'B3': '14', 'B4': '13', 'B5': '12', 'B6': '11', 'B7': '10', 'B8': '09',
        'C1': '17', 'C2': '18', 'C3': '19', 'C4': '20', 'C5': '21', 'C6': '22', 'C7': '23', 'C8': '24',
        'D1': '32', 'D2': '31', 'D3': '30', 'D4': '29', 'D5': '28', 'D6': '27', 'D7': '26', 'D8': '25',
        'E1': '33', 'E2': '34', 'E3': '35', 'E4': '36', 'E5': '37', 'E6': '38', 'E7': '39', 'E8': '40',
        'F1': '48', 'F2': '47', 'F3': '46', 'F4': '45', 'F5': '44', 'F6': '43', 'F7': '42', 'F8': '41',
        'G1': '49', 'G2': '50', 'G3': '51', 'G4': '52', 'G5': '53', 'G6': '54', 'G7': '55', 'G8': '56',
        'H1': '64', 'H2': '63', 'H3': '62', 'H4': '61', 'H5': '60', 'H6': '59', 'H7': '58', 'H8': '57',
    }
    # fmt: on
    chip_type = int(caget(pv.me14e_gp1))
    if chip_type in [ChipType.Oxford, ChipType.OxfordInner]:
        logger.info("Oxford Block Order")
        rows = ["A", "B", "C", "D", "E", "F", "G", "H"]
        columns = list(range(1, 10))
        btn_names = {}
        flip = True
        for x, column in enumerate(columns):
            for y, row in enumerate(rows):
                i = x * 8 + y
                if i % 8 == 0 and flip is False:
                    flip = True
                    z = 8 - (y + 1)
                elif i % 8 == 0 and flip is True:
                    flip = False
                    z = y
                elif flip is False:
                    z = y
                elif flip is True:
                    z = 8 - (y + 1)
                else:
                    logger.warning("Problem in Chip Grid Creation")
                    break
                button_name = str(row) + str(column)
                lab_num = x * 8 + z
                label = "%02.d" % (lab_num + 1)
                btn_names[button_name] = label
        block_dict = btn_names

    litemap_path = _coerce_to_path(litemap_path)

    litemap_fid = str(caget(pv.me14e_gp5)) + ".lite"
    logger.info("Please wait, loading LITE map")
    logger.info("Loading Lite Map")
    logger.info("Opening %s" % (litemap_path / litemap_fid))
    with open(litemap_path / litemap_fid, "r") as fh:
        f = fh.readlines()
    for line in f:
        entry = line.split()
        block_name = entry[0]
        yesno = entry[1]
        block_num = block_dict[block_name]
        pvar = "ME14E-MO-IOC-01:GP" + str(int(block_num) + 10)
        logger.info("Block: %s \tScanned: %s \tPVAR: %s" % (block_name, yesno, pvar))
    logger.debug("Load lite map done")


@log.log_on_entry
def load_full_map(fullmap_path: Path | str = FULLMAP_PATH):
    (
        chip_name,
        visit,
        sub_dir,
        n_exposures,
        chip_type,
        map_type,
    ) = startup.scrape_parameter_file()
    fullmap_path = _coerce_to_path(fullmap_path)

    fullmap_fid = fullmap_path / f"{str(caget(pv.me14e_gp5))}.spec"
    logger.info("Opening %s" % fullmap_fid)
    mapping.plot_file(fullmap_fid, chip_type)
    mapping.convert_chip_to_hex(fullmap_fid, chip_type)
    shutil.copy2(fullmap_fid.with_suffix(".full"), fullmap_path / "currentchip.full")
    logger.info(
        "Copying %s to %s"
        % (fullmap_fid.with_suffix(".full"), fullmap_path / "currentchip.full")
    )
    logger.debug("Load full map done")


@log.log_on_entry
def moveto(place: str = "origin"):
    logger.info("Move to: %s" % place)
    chip_type = int(caget(pv.me14e_gp1))
    logger.info("Chip type is%s" % chip_type)

    if chip_type == ChipType.Oxford or chip_type == ChipType.Minichip:
        # Oxford and minichip
        # As minichip is nothing more than a smaller oxford,
        # they should move the same way
        logger.info("Oxford Move")
        if place == "origin":
            caput(pv.me14e_stage_x, 0.0)
            caput(pv.me14e_stage_y, 0.0)
        if place == "f1":
            caput(pv.me14e_stage_x, 25.40)
            caput(pv.me14e_stage_y, 0.0)
        if place == "f2":
            caput(pv.me14e_stage_x, 0.0)
            caput(pv.me14e_stage_y, 25.40)

    elif chip_type == ChipType.OxfordInner:
        logger.info("Oxford Inner Move")
        if place == "origin":
            caput(pv.me14e_stage_x, 0.0)
            caput(pv.me14e_stage_y, 0.0)
        if place == "f1":
            caput(pv.me14e_stage_x, 24.60)
            caput(pv.me14e_stage_y, 0.0)
        if place == "f2":
            caput(pv.me14e_stage_x, 0.0)
            caput(pv.me14e_stage_y, 24.60)

    elif chip_type == ChipType.Custom:
        logger.info("Custom Chip Move")
        if place == "origin":
            caput(pv.me14e_stage_x, 0.0)
            caput(pv.me14e_stage_y, 0.0)
        if place == "f1":
            caput(pv.me14e_stage_x, 25.40)
            caput(pv.me14e_stage_y, 0.0)
        if place == "f2":
            caput(pv.me14e_stage_x, 0.0)
            caput(pv.me14e_stage_y, 25.40)

    else:
        logger.warning("Unknown chip_type move")

    # Non Chip Specific Move
    if place == "zero":
        logger.info("Moving to %s" % place)
        caput(pv.me14e_pmac_str, "!x0y0z0")

    elif place == "load_position":
        logger.info("load position")
        caput(pv.bs_mp_select, "Robot")
        caput(pv.bl_mp_select, "Out")
        caput(pv.det_z, 1300)

    elif place == "collect_position":
        logger.info("collect position")
        caput(pv.me14e_filter, 20)
        caput(pv.me14e_stage_x, 0.0)
        caput(pv.me14e_stage_y, 0.0)
        caput(pv.me14e_stage_z, 0.0)
        caput(pv.bs_mp_select, "Data Collection")
        caput(pv.bl_mp_select, "In")

    elif place == "microdrop_position":
        logger.info("microdrop align position")
        caput(pv.me14e_stage_x, 6.0)
        caput(pv.me14e_stage_y, -7.8)
        caput(pv.me14e_stage_z, 0.0)

    elif place == "laser1on":  # these are in laser edm
        logger.info("Laser 1 /BNC2 shutter is open")
        # Use M712 = 0 if triggering on falling edge. M712 =1 if on rising edge
        # Be sure to also change laser1off
        # caput(pv.me14e_pmac_str, ' M712=0 M711=1')
        caput(pv.me14e_pmac_str, " M712=1 M711=1")

    elif place == "laser1off":
        logger.info("Laser 1 shutter is closed")
        caput(pv.me14e_pmac_str, " M712=0 M711=1")

    elif place == "laser2on":
        logger.info("Laser 2 / BNC3 shutter is open")
        caput(pv.me14e_pmac_str, " M812=1 M811=1")

    elif place == "laser2off":
        logger.info("Laser 2 shutter is closed")
        caput(pv.me14e_pmac_str, " M812=0 M811=1")

    elif place == "laser1burn":
        led_burn_time = caget(pv.me14e_gp103)
        logger.info("Laser 1  on")
        logger.info("Burn time is %s s" % led_burn_time)
        caput(pv.me14e_pmac_str, " M712=1 M711=1")
        sleep(int(float(led_burn_time)))
        logger.info("Laser 1 off")
        caput(pv.me14e_pmac_str, " M712=0 M711=1")

    elif place == "laser2burn":
        led_burn_time = caget(pv.me14e_gp109)
        logger.info("Laser 2 on")
        logger.info("burntime %s s" % led_burn_time)
        caput(pv.me14e_pmac_str, " M812=1 M811=1")
        sleep(int(float(led_burn_time)))
        logger.info("Laser 2 off")
        caput(pv.me14e_pmac_str, " M812=0 M811=1")


@log.log_on_entry
def scrape_mtr_directions(param_path: Path | str = CS_FILES_PATH):
    param_path = _coerce_to_path(param_path)

    with open(param_path / "motor_direction.txt", "r") as f:
        lines = f.readlines()
    mtr1_dir, mtr2_dir, mtr3_dir = 1.0, 1.0, 1.0
    for line in lines:
        if line.startswith("mtr1"):
            mtr1_dir = float(line.split("=")[1])
        elif line.startswith("mtr2"):
            mtr2_dir = float(line.split("=")[1])
        elif line.startswith("mtr3"):
            mtr3_dir = float(line.split("=")[1])
        else:
            continue
    logger.info("mt1_dir %s mtr2_dir %s mtr3_dir %s" % (mtr1_dir, mtr2_dir, mtr3_dir))
    return mtr1_dir, mtr2_dir, mtr3_dir


@log.log_on_entry
def fiducial(point: int = 1, param_path: Path | str = CS_FILES_PATH):
    scale = 10000.0  # noqa: F841
    param_path = _coerce_to_path(param_path)

    mtr1_dir, mtr2_dir, mtr3_dir = scrape_mtr_directions(param_path)

    rbv_1 = float(caget(pv.me14e_stage_x + ".RBV"))
    rbv_2 = float(caget(pv.me14e_stage_y + ".RBV"))
    rbv_3 = float(caget(pv.me14e_stage_z + ".RBV"))

    raw_1 = float(caget(pv.me14e_stage_x + ".RRBV"))
    raw_2 = float(caget(pv.me14e_stage_y + ".RRBV"))
    raw_3 = float(caget(pv.me14e_stage_z + ".RRBV"))

    f_x = rbv_1
    f_y = rbv_2
    f_z = rbv_3

    logger.info("Writing Fiducial File %s/fiducial_%s.txt" % (param_path, point))
    logger.info("MTR\tRBV\tRAW\tCorr\tf_value")
    logger.info("MTR1\t%1.4f\t%i\t%i\t%1.4f" % (rbv_1, raw_1, mtr1_dir, f_x))
    logger.info("MTR2\t%1.4f\t%i\t%i\t%1.4f" % (rbv_2, raw_2, mtr2_dir, f_y))
    logger.info("MTR3\t%1.4f\t%i\t%i\t%1.4f" % (rbv_3, raw_3, mtr3_dir, f_z))

    with open(param_path / f"fiducial_{point}.txt", "w") as f:
        f.write("MTR\tRBV\tRAW\tCorr\tf_value\n")
        f.write("MTR1\t%1.4f\t%i\t%i\t%1.4f\n" % (rbv_1, raw_1, mtr1_dir, f_x))
        f.write("MTR2\t%1.4f\t%i\t%i\t%1.4f\n" % (rbv_2, raw_2, mtr2_dir, f_y))
        f.write("MTR3\t%1.4f\t%i\t%i\t%1.4f" % (rbv_3, raw_3, mtr3_dir, f_z))
    logger.info(f"Fiducial {point} set.")


def scrape_mtr_fiducials(point: int, param_path: Path | str = CS_FILES_PATH):
    param_path = _coerce_to_path(param_path)

    with open(param_path / f"fiducial_{point}.txt", "r") as f:
        f_lines = f.readlines()[1:]
    f_x = float(f_lines[0].rsplit()[4])
    f_y = float(f_lines[1].rsplit()[4])
    f_z = float(f_lines[2].rsplit()[4])
    return f_x, f_y, f_z


@log.log_on_entry
def cs_maker():
    """
    Coordinate system.

    Values for scalex, scaley, scalez, and skew, as well as the sign of
    Sx, Sy, Sz are stored in a .json file and should be modified there.
    Location of file: src/mx_bluesky/I24/serial/parameters/cs_maker.json

    Theory
    Rx: rotation about X-axis, pitch
    Ry: rotation about Y-axis, yaw
    Rz: rotation about Z-axis, roll
    The order of rotation is Roll->Yaw->Pitch (Rx*Ry*Rz)
    Rx           Ry          Rz
    |1  0   0| | Cy  0 Sy| |Cz -Sz 0|   | CyCz        -CxSz         Sy  |
    |0 Cx -Sx|*|  0  1  0|*|Sz  Cz 0| = | SxSyCz+CxSz -SxSySz+CxCz -SxCy|
    |0 Sx  Cx| |-Sy  0 Cy| | 0   0 1|   |-CxSyCz+SxSz  CxSySz+SxCz  CxCy|

    BELOW iS TEST TEST (CLOCKWISE)
    Rx           Ry          Rz
    |1  0   0| | Cy 0 -Sy| |Cz  Sz 0|   | CyCz         CxSz         -Sy |
    |0 Cx  Sx|*|  0  1  0|*|-Sz Cz 0| = | SxSyCz-CxSz  SxSySz+CxCz  SxCy|
    |0 -Sx Cx| | Sy  0 Cy| | 0   0 1|   | CxSyCz+SxSz  CxSySz-SxCz  CxCy|


    Skew:
    Skew is the difference between the Sz1 and Sz2 after rotation is taken out.
    This should be measured in situ prior to expriment, ie. measure by hand using
    opposite and adjacent RBV after calibration of scale factors.
    """
    chip_type = int(caget(pv.me14e_gp1))
    fiducial_dict = {}
    fiducial_dict[0] = [25.400, 25.400]
    fiducial_dict[1] = [24.600, 24.600]
    fiducial_dict[2] = [25.400, 25.400]
    fiducial_dict[3] = [18.25, 18.25]
    logger.info("Chip type is %s with size %s" % (chip_type, fiducial_dict[chip_type]))

    mtr1_dir, mtr2_dir, mtr3_dir = scrape_mtr_directions()
    f1_x, f1_y, f1_z = scrape_mtr_fiducials(1)
    f2_x, f2_y, f2_z = scrape_mtr_fiducials(2)
    logger.info("mtr1 direction: %s" % mtr1_dir)
    logger.info("mtr2 direction: %s" % mtr2_dir)
    logger.info("mtr3 direction: %s" % mtr3_dir)

    # Scale parameters saved in json file
    try:
        with open(CS_FILES_PATH / "cs_maker.json", "r") as fh:
            cs_info = json.load(fh)
    except json.JSONDecodeError:
        logger.error("Invalid JSON file.")
        raise

    try:
        scalex, scaley, scalez = (
            float(cs_info["scalex"]),
            float(cs_info["scaley"]),
            float(cs_info["scalez"]),
        )
        skew = float(cs_info["skew"])
        Sx_dir, Sy_dir, Sz_dir = (
            int(cs_info["Sx_dir"]),
            int(cs_info["Sy_dir"]),
            int(cs_info["Sz_dir"]),
        )
    except KeyError:
        logger.error("Wrong or missing key in the cs json file.")
        raise

    def check_dir(val):
        if val not in [1, -1]:
            raise ValueError("Wrong value for direction. Please set to either -1 or 1.")

    check_dir(Sx_dir)
    check_dir(Sy_dir)
    check_dir(Sz_dir)

    # Rotation Around Z
    # If stages upsidedown (I24) change sign of Sz
    Sz1 = -1 * f1_y / fiducial_dict[chip_type][0]
    Sz2 = f2_x / fiducial_dict[chip_type][1]
    Sz = Sz_dir * ((Sz1 + Sz2) / 2)
    Cz = np.sqrt((1 - Sz**2))
    logger.info("Sz1 , %1.4f, %1.4f" % (Sz1, np.degrees(np.arcsin(Sz1))))
    logger.info("Sz2 , %1.4f, %1.4f" % (Sz2, np.degrees(np.arcsin(Sz2))))
    logger.info("Sz , %1.4f, %1.4f" % (Sz, np.degrees(np.arcsin(Sz))))
    logger.info("Cz , %1.4f, %1.4f" % (Cz, np.degrees(np.arcsin(Cz))))
    # Rotation Around Y
    Sy = Sy_dir * f1_z / fiducial_dict[chip_type][0]
    Cy = np.sqrt((1 - Sy**2))
    logger.info("Sy , %1.4f, %1.4f" % (Sy, np.degrees(np.arcsin(Sy))))
    logger.info("Cy , %1.4f, %1.4f" % (Cy, np.degrees(np.arcsin(Cy))))
    # Rotation Around X
    # If stages upsidedown (I24) change sign of Sx
    Sx = Sx_dir * f2_z / fiducial_dict[chip_type][1]
    Cx = np.sqrt((1 - Sx**2))
    logger.info("Sx , %1.4f, %1.4f" % (Sx, np.degrees(np.arcsin(Sx))))
    logger.info("Cx , %1.4f, %1.4f" % (Cx, np.degrees(np.arcsin(Cx))))

    x1factor = mtr1_dir * scalex * (Cy * Cz)
    y1factor = mtr2_dir * scaley * (-1.0 * Cx * Sz)
    z1factor = mtr3_dir * scalez * Sy

    x2factor = mtr1_dir * scalex * ((Sx * Sy * Cz) + (Cx * Sz))
    y2factor = mtr2_dir * scaley * ((Cx * Cz) - (Sx * Sy * Sz))
    z2factor = mtr3_dir * scalez * (-1.0 * Sx * Cy)

    x3factor = mtr1_dir * scalex * ((Sx * Sz) - (Cx * Sy * Cz))
    y3factor = mtr2_dir * scaley * ((Cx * Sy * Sz) + (Sx * Cz))
    z3factor = mtr3_dir * scalez * (Cx * Cy)

    logger.info("Skew being used is: %1.4f" % skew)
    s1 = np.degrees(np.arcsin(Sz1))
    s2 = np.degrees(np.arcsin(Sz2))
    rot = np.degrees(np.arcsin((Sz1 + Sz2) / 2))
    calc_skew = (s1 - rot) - (s2 - rot)
    logger.info("s1:%1.4f s2:%1.4f rot:%1.4f" % (s1, s2, rot))
    logger.info("Calculated rotation from current fiducials is: %1.4f" % rot)
    logger.info("Calculated Skew from current fiducials is: %1.4f" % calc_skew)
    logger.info("Calculated Skew has been known to have the wrong sign")

    sinD = np.sin((skew / 2) * (np.pi / 180))
    cosD = np.cos((skew / 2) * (np.pi / 180))
    new_x1factor = (x1factor * cosD) + (y1factor * sinD)
    new_y1factor = (x1factor * sinD) + (y1factor * cosD)
    new_x2factor = (x2factor * cosD) + (y2factor * sinD)
    new_y2factor = (x2factor * sinD) + (y2factor * cosD)

    cs1 = "#1->%+1.3fX%+1.3fY%+1.3fZ" % (new_x1factor, new_y1factor, z1factor)
    cs2 = "#2->%+1.3fX%+1.3fY%+1.3fZ" % (new_x2factor, new_y2factor, z2factor)
    cs3 = "#3->%+1.3fX%+1.3fY%+1.3fZ" % (x3factor, y3factor, z3factor)
    logger.info("PMAC strings. \ncs1: %s \ncs2: %scs3: %s" % (cs1, cs2, cs3))
    logger.info(
        """These next values should be 1.
        This is the sum of the squares of the factors divided by their scale."""
    )
    sqfact1 = np.sqrt(x1factor**2 + y1factor**2 + z1factor**2) / scalex
    sqfact2 = np.sqrt(x2factor**2 + y2factor**2 + z2factor**2) / scaley
    sqfact3 = np.sqrt(x3factor**2 + y3factor**2 + z3factor**2) / scalez
    logger.info("%1.4f \n %1.4f \n %1.4f" % (sqfact1, sqfact2, sqfact3))
    logger.info("Long wait, please be patient")
    caput(pv.me14e_pmac_str, "!x0y0z0")
    sleep(2.5)
    caput(pv.me14e_pmac_str, "&2")
    caput(pv.me14e_pmac_str, cs1)
    caput(pv.me14e_pmac_str, cs2)
    caput(pv.me14e_pmac_str, cs3)
    caput(pv.me14e_pmac_str, "!x0y0z0")
    sleep(0.1)
    caput(pv.me14e_pmac_str, "#1hmz#2hmz#3hmz")
    sleep(0.1)
    logger.info("Chip_type is %s" % chip_type)
    if chip_type == 0:
        caput(pv.me14e_pmac_str, "!x0.4y0.4")
        sleep(0.1)
        caput(pv.me14e_pmac_str, "#1hmz#2hmz#3hmz")
    else:
        caput(pv.me14e_pmac_str, "#1hmz#2hmz#3hmz")
    logger.debug("CSmaker done.")


def cs_reset():
    cs1 = "#1->-10000X+0Y+0Z"
    cs2 = "#2->+0X+10000Y+0Z"
    cs3 = "#3->0X+0Y+10000Z"
    strg = "\n".join([cs1, cs2, cs3])
    print(strg)
    caput(pv.me14e_pmac_str, "&2")
    sleep(0.5)
    caput(pv.me14e_pmac_str, cs1)
    sleep(0.5)
    caput(pv.me14e_pmac_str, cs2)
    sleep(0.5)
    caput(pv.me14e_pmac_str, cs3)
    logger.debug("CSreset Done")


@log.log_on_entry
def pumpprobe_calc():
    logger.info("Calculate and show exposure and dwell time for each option.")
    exptime = float(caget(pv.me14e_exptime))
    pumpexptime = float(caget(pv.me14e_gp103))
    movetime = 0.008
    logger.info("X-ray exposure time %s" % exptime)
    logger.info("Laser dwell time %s" % pumpexptime)
    repeat1 = 2 * 20 * (movetime + (pumpexptime + exptime) / 2)
    repeat2 = 4 * 20 * (movetime + (pumpexptime + exptime) / 2)
    repeat3 = 6 * 20 * (movetime + (pumpexptime + exptime) / 2)
    repeat5 = 10 * 20 * (movetime + (pumpexptime + exptime) / 2)
    repeat10 = 20 * 20 * (movetime + (pumpexptime + exptime) / 2)
    for pv_name, repeat in (
        (pv.me14e_gp104, repeat1),
        (pv.me14e_gp105, repeat2),
        (pv.me14e_gp106, repeat3),
        (pv.me14e_gp107, repeat5),
        (pv.me14e_gp108, repeat10),
    ):
        rounded = round(repeat, 4)
        caput(pv_name, rounded)
        logger.info("Repeat (%s): %s s" % (pv_name, rounded))
    # logger.info("repeat10 (%s): %s s" % (pv.me14e_gp108, round(repeat10, 4)))
    logger.debug("PP calculations done")


@log.log_on_entry
def block_check():
    caput(pv.me14e_gp9, 0)
    while True:
        if int(caget(pv.me14e_gp9)) == 0:
            chip_type = int(caget(pv.me14e_gp1))
            if chip_type == ChipType.Minichip:
                logger.info("Oxford mini chip in use.")
                block_start_list = scrape_pvar_file("minichip_oxford.pvar")
            elif chip_type == ChipType.Custom:
                logger.error("This is a custom chip, no block check available!")
                raise ValueError(
                    "Chip type set to 'custom', which has no block check."
                    "If not using a custom chip, please double check chip in the GUI."
                )
            else:
                logger.warning("Default is Oxford chip block start list.")
                block_start_list = scrape_pvar_file("oxford.pvar")
            for entry in block_start_list:
                if int(caget(pv.me14e_gp9)) != 0:
                    logger.warning("Block Check Aborted")
                    sleep(1.0)
                    break
                block, x, y = entry
                logger.info("Block: %s -> (x=%s y=%s)" % (block, x, y))
                caput(pv.me14e_pmac_str, "!x%sy%s" % (x, y))
                time.sleep(0.4)
        else:
            logger.warning("Block Check Aborted due to GP 9 not equalling 0")
            break
        break
    logger.debug("Block check done")


def parse_args_and_run_parsed_function(args):
    print(f"Run with {args}")
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(
        help="Choose command.",
        required=True,
        dest="sub-command",
    )
    parser_init = subparsers.add_parser(
        "initialise",
    )
    parser_init.set_defaults(func=initialise)
    parser_moveto = subparsers.add_parser(
        "moveto",
    )
    parser_moveto.add_argument("place", type=str)
    parser_moveto.set_defaults(func=moveto)
    parser_fid = subparsers.add_parser("fiducial")
    parser_fid.add_argument("point", type=int)
    parser_fid.set_defaults(func=fiducial)
    parser_csm = subparsers.add_parser("cs_maker")
    parser_csm.set_defaults(func=cs_maker)
    parser_pp = subparsers.add_parser("pumpprobe_calc")
    parser_pp.set_defaults(func=pumpprobe_calc)
    parser_write = subparsers.add_parser("write_parameter_file")
    parser_write.set_defaults(func=write_parameter_file)
    parser_def = subparsers.add_parser("define_current_chip")
    parser_def.add_argument("chipid", type=str)
    parser_def.set_defaults(func=define_current_chip)
    parser_stockmap = subparsers.add_parser("load_stock_map")
    parser_stockmap.add_argument("map_choice", type=str)
    parser_stockmap.set_defaults(func=load_stock_map)
    parser_litemap = subparsers.add_parser("load_lite_map")
    parser_litemap.set_defaults(func=load_lite_map)
    parser_fullmap = subparsers.add_parser("load_full_map")
    parser_fullmap.set_defaults(func=load_full_map)
    parser_save = subparsers.add_parser("save_screen_map")
    parser_save.set_defaults(func=save_screen_map)
    parser_upld = subparsers.add_parser("upload_full")
    parser_upld.set_defaults(func=upload_full)
    parser_params = subparsers.add_parser("upload_parameters")
    parser_params.add_argument("chipid", type=str)
    parser_params.set_defaults(func=upload_parameters)
    parser_csr = subparsers.add_parser("cs_reset")
    parser_csr.set_defaults(func=cs_reset)
    parser_block = subparsers.add_parser("block_check")
    parser_block.set_defaults(func=block_check)

    args = parser.parse_args(args)

    args_dict = vars(args)
    args_dict.pop("sub-command")
    args_dict.pop("func")(**args_dict)


if __name__ == "__main__":
    setup_logging()

    parse_args_and_run_parsed_function(sys.argv[1:])
